from datetime import datetime
from typing import Any, Dict, List, Optional

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    Float,
    ForeignKey,
    Index,
    Integer,
    String,
    Table,
    Text,
    UniqueConstraint,
    func,
    orm,
    text,
)
from sqlalchemy.dialects import postgresql as pg

from balsam.schemas.transfer import TransferDirection, TransferItemState

from .base import Base

# PK automatically has nullable=False, autoincrement
# Postgres auto-creates index for unique constraint and primary key constraint


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    username = Column(String(100), unique=True)
    hashed_password = Column(String(128), nullable=True, default=None)


class DeviceCodeAttempt(Base):
    __tablename__ = "device_code_attempts"

    client_id = Column(pg.UUID(as_uuid=True), primary_key=True)
    expiration = Column(DateTime, nullable=False, default=datetime.utcnow)
    device_code = Column(String(1024))
    user_code = Column(String(16), unique=True)
    scope = Column(String(128))
    user_denied = Column(Boolean, default=False)
    user_id = Column(Integer, ForeignKey("users.id", ondelete="CASCADE"), nullable=True, default=None)
    user = orm.relationship(User)


class AuthorizationState(Base):
    __tablename__ = "auth_states"
    id = Column(Integer, primary_key=True)
    state = Column(String(512), unique=True)


class Site(Base):
    __tablename__ = "sites"
    __table_args__ = (UniqueConstraint("owner_id", "name"),)

    id = Column(Integer, primary_key=True)
    name = Column(String(100))
    path = Column(String(512))
    last_refresh = Column(DateTime)
    creation_date = Column(DateTime, default=datetime.utcnow)
    owner_id = Column(Integer, ForeignKey("users.id", ondelete="CASCADE"), index=True)
    globus_endpoint_id = Column(pg.UUID(as_uuid=True))

    backfill_windows = Column(JSON, default=list)
    queued_jobs = Column(JSON, default=list)
    optional_batch_job_params: Dict[str, Any] = Column(JSON, default=dict)  # type: ignore
    allowed_projects = Column(JSON, default=list)
    allowed_queues: Dict[str, Any] = Column(JSON, default=dict)  # type: ignore
    transfer_locations: Dict[str, Any] = Column(pg.JSONB, default=dict)  # type: ignore

    owner = orm.relationship(User, lazy="raise")
    apps = orm.relationship(
        "App",
        back_populates="site",
        lazy="dynamic",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    batch_jobs = orm.relationship(
        "BatchJob",
        back_populates="site",
        lazy="dynamic",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    sessions = orm.relationship(
        "Session",
        back_populates="site",
        lazy="dynamic",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


class App(Base):
    __tablename__ = "apps"
    __table_args__ = (UniqueConstraint("site_id", "name"),)

    id = Column(Integer, primary_key=True)
    site_id = Column(Integer, ForeignKey("sites.id", ondelete="CASCADE"))
    name = Column(String(200), nullable=False)
    description = Column(Text)
    parameters: Dict[str, Any] = Column(JSON, default=dict)  # type: ignore
    transfers: Dict[str, Any] = Column(JSON, default=dict)  # type: ignore
    serialized_class = Column(Text, default="")
    source_code = Column(Text, default="")

    site = orm.relationship(Site, back_populates="apps")
    jobs = orm.relationship("Job", back_populates="app", cascade="all, delete-orphan", passive_deletes=True)


# Junction table
job_deps = Table(
    "job_deps",
    Base.metadata,
    Column(
        "parent_id",
        ForeignKey("jobs.id", ondelete="CASCADE"),
        index=True,
        primary_key=True,
    ),
    Column(
        "child_id",
        ForeignKey("jobs.id", ondelete="CASCADE"),
        index=True,
        primary_key=True,
    ),
)


class Job(Base):
    __tablename__ = "jobs"

    id = Column(Integer, primary_key=True, index=True)
    workdir = Column(String(256), nullable=False)

    # https://www.postgresql.org/docs/9.4/datatype-json.html
    # We want tag-based queries to test for containment
    #     SELECT ... WHERE jobs.tags @> '{"myKey": "myValue"}'::jsonb
    # Ensure that GIN index is created:
    #     CREATE INDEX idxgin ON jobs USING gin (tags);
    tags = Column(pg.JSONB, default=dict, index=True)
    __table_args__ = (Index("ix_jobs_tags", text("(tags jsonb_path_ops)"), postgresql_using="GIN"),)

    app_id = Column(Integer, ForeignKey("apps.id", ondelete="CASCADE"))
    session_id = Column(
        Integer,
        ForeignKey("sessions.id", ondelete="SET NULL"),
        nullable=True,
        default=None,
    )
    serialized_parameters = Column(Text, default="")
    serialized_return_value = Column(Text, default="")
    serialized_exception = Column(Text, default="")
    batch_job_id = Column(Integer, ForeignKey("batch_jobs.id", ondelete="SET NULL"), nullable=True)
    state = Column(String(32), index=True)
    last_update = Column(DateTime(timezone=True), default=func.now(), onupdate=func.now())
    data = Column(JSON)
    return_code = Column(Integer)
    pending_file_cleanup = Column(Boolean, default=True)

    parents = orm.relationship(
        "Job",
        secondary=job_deps,
        primaryjoin=(id == job_deps.c.child_id),
        secondaryjoin=(id == job_deps.c.parent_id),
        back_populates="children",
    )
    parent_ids: Optional[List[int]]
    children = orm.relationship(
        "Job",
        secondary=job_deps,
        primaryjoin=(id == job_deps.c.parent_id),
        secondaryjoin=(id == job_deps.c.child_id),
        back_populates="parents",
        lazy="raise",
    )

    num_nodes = Column(Integer)
    ranks_per_node = Column(Integer)
    threads_per_rank = Column(Integer)
    threads_per_core = Column(Integer)
    gpus_per_rank = Column(Float)
    node_packing_count = Column(Integer)
    wall_time_min = Column(Integer)
    launch_params = Column(JSON)

    app = orm.relationship("App", back_populates="jobs")
    session: Optional["Session"] = orm.relationship("Session", back_populates="jobs")  # type: ignore
    batch_job = orm.relationship("BatchJob", back_populates="jobs")
    transfer_items: List["TransferItem"] = orm.relationship(
        "TransferItem",
        lazy="raise",
        back_populates="job",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )  # type: ignore
    log_events = orm.relationship(
        "LogEvent",
        lazy="raise",
        back_populates="job",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


class BatchJob(Base):
    __tablename__ = "batch_jobs"

    id = Column(Integer, primary_key=True)
    site_id = Column(Integer, ForeignKey("sites.id", ondelete="CASCADE"), nullable=False)
    scheduler_id = Column(Integer, nullable=True)
    project = Column(String(64), nullable=False)
    queue = Column(String(64), nullable=False)
    optional_params = Column(JSON, default=dict)
    num_nodes = Column(Integer, nullable=False)
    wall_time_min = Column(Integer, nullable=False)
    job_mode = Column(String(16), nullable=False)
    partitions = Column(JSON, default=None, nullable=True)
    filter_tags = Column(pg.JSONB, default=dict)
    state = Column(String(32), index=True, nullable=False)
    status_info = Column(JSON, default=dict)
    start_time = Column(DateTime, nullable=True)
    end_time = Column(DateTime, nullable=True)

    site = orm.relationship(Site, back_populates="batch_jobs")
    jobs = orm.relationship("Job", back_populates="batch_job")
    sessions = orm.relationship("Session", back_populates="batch_job")


class Session(Base):
    __tablename__ = "sessions"

    id = Column(Integer, primary_key=True)
    heartbeat = Column(DateTime, default=datetime.utcnow)
    batch_job_id = Column(Integer, ForeignKey("batch_jobs.id", ondelete="SET NULL"), nullable=True)
    site_id = Column(Integer, ForeignKey("sites.id", ondelete="CASCADE"), nullable=False)

    batch_job = orm.relationship(BatchJob, lazy="raise", back_populates="sessions")
    site = orm.relationship(Site, lazy="raise", back_populates="sessions")
    jobs: "orm.Query[Job]" = orm.relationship("Job", lazy="dynamic", back_populates="session")  # type: ignore


class TransferItem(Base):
    __tablename__ = "transfer_items"

    id = Column(Integer, primary_key=True)
    job_id = Column(Integer, ForeignKey("jobs.id", ondelete="CASCADE"), nullable=False)
    direction = Column(Enum(TransferDirection), nullable=False)
    local_path = Column(String(256))
    remote_path = Column(String(256))
    recursive = Column(Boolean, default=False)
    location_alias = Column(String(256))
    state = Column(Enum(TransferItemState), nullable=False)
    task_id = Column(String(100))
    transfer_info = Column(JSON, default=dict)

    job = orm.relationship(Job, back_populates="transfer_items")


class LogEvent(Base):
    __tablename__ = "log_events"

    id = Column(Integer, primary_key=True)
    job_id = Column(Integer, ForeignKey("jobs.id", ondelete="CASCADE"))
    timestamp = Column(DateTime)
    from_state = Column(String(32))
    to_state = Column(String(32))
    data = Column(pg.JSONB, default=dict)

    job = orm.relationship(Job, back_populates="log_events")



# This file was auto-generated via /Users/misha/workflow/balsam/.venv/bin/python balsam/schemas/api_generator.py
# [git rev 4b3ff11]
# Do *not* make changes to the API by changing this file!

import datetime
import pathlib
import typing
import uuid
from typing import Any, List, Optional, Union

import pydantic

import balsam._api.bases
import balsam._api.model
from balsam._api.model import Field
from balsam._api.query import Query


class Site(balsam._api.bases.SiteBase):
    _create_model_cls = balsam.schemas.site.SiteCreate
    _update_model_cls = balsam.schemas.site.SiteUpdate
    _read_model_cls = balsam.schemas.site.SiteOut
    objects: "SiteManager"

    name = Field[str]()
    path = Field[pathlib.Path]()
    globus_endpoint_id = Field[Optional[uuid.UUID]]()
    backfill_windows = Field[typing.Dict[str, typing.List[balsam.schemas.batchjob.SchedulerBackfillWindow]]]()
    queued_jobs = Field[typing.Dict[int, balsam.schemas.batchjob.SchedulerJobStatus]]()
    optional_batch_job_params = Field[typing.Dict[str, str]]()
    allowed_projects = Field[typing.List[str]]()
    allowed_queues = Field[typing.Dict[str, balsam.schemas.site.AllowedQueue]]()
    transfer_locations = Field[typing.Dict[str, pydantic.networks.AnyUrl]]()
    id = Field[Optional[int]]()
    last_refresh = Field[Optional[datetime.datetime]]()
    creation_date = Field[Optional[datetime.datetime]]()

    def __init__(
        self,
        name: str,
        path: pathlib.Path,
        globus_endpoint_id: Optional[uuid.UUID] = None,
        backfill_windows: Optional[
            typing.Dict[str, typing.List[balsam.schemas.batchjob.SchedulerBackfillWindow]]
        ] = None,
        queued_jobs: Optional[typing.Dict[int, balsam.schemas.batchjob.SchedulerJobStatus]] = None,
        optional_batch_job_params: Optional[typing.Dict[str, str]] = None,
        allowed_projects: Optional[typing.List[str]] = None,
        allowed_queues: Optional[typing.Dict[str, balsam.schemas.site.AllowedQueue]] = None,
        transfer_locations: Optional[typing.Dict[str, pydantic.networks.AnyUrl]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new Site object.  You must eventually call the save() method or
        pass a Site list into Site.objects.bulk_create().

        name:                      Unique Site name
        path:                      Absolute filesystem path of the Site
        globus_endpoint_id:        Associated Globus endpoint ID
        backfill_windows:          Idle backfill currently available at the Site, keyed by queue name
        queued_jobs:               Queued scheduler jobs at the Site, keyed by scheduler ID
        optional_batch_job_params: Optional pass-through parameters accepted by the Site batchjob template
        allowed_projects:          Allowed projects/allocations for batchjob submission
        allowed_queues:            Allowed queues and associated queueing policies
        transfer_locations:        Trusted transfer location aliases and associated protocol/URLs
        """
        _kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        _kwargs.update(kwargs)
        return super().__init__(**_kwargs)


class SiteQuery(Query[Site]):
    def get(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        id: Union[typing.List[int], int, None] = None,
        last_refresh_after: Optional[datetime.datetime] = None,
    ) -> Site:
        """
        Retrieve exactly one Site. Raises Site.DoesNotExist
        if no items were found, or Site.MultipleObjectsReturned if
        more than one item matched the query.

        name:               Fetch the site with this name
        path:               Only return Sites with paths containing this string.
        id:                 Only return Sites having an id in this list.
        last_refresh_after: Only return Sites active since this time (UTC)
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        id: Union[typing.List[int], int, None] = None,
        last_refresh_after: Optional[datetime.datetime] = None,
    ) -> "SiteQuery":
        """
        Retrieve exactly one Site. Raises Site.DoesNotExist
        if no items were found, or Site.MultipleObjectsReturned if
        more than one item matched the query.

        name:               Fetch the site with this name
        path:               Only return Sites with paths containing this string.
        id:                 Only return Sites having an id in this list.
        last_refresh_after: Only return Sites active since this time (UTC)
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)

    def update(
        self,
        name: Optional[str] = None,
        path: Optional[pathlib.Path] = None,
        globus_endpoint_id: Optional[uuid.UUID] = None,
        backfill_windows: Optional[
            typing.Dict[str, typing.List[balsam.schemas.batchjob.SchedulerBackfillWindow]]
        ] = None,
        queued_jobs: Optional[typing.Dict[int, balsam.schemas.batchjob.SchedulerJobStatus]] = None,
        optional_batch_job_params: Optional[typing.Dict[str, str]] = None,
        allowed_projects: Optional[typing.List[str]] = None,
        allowed_queues: Optional[typing.Dict[str, balsam.schemas.site.AllowedQueue]] = None,
        transfer_locations: Optional[typing.Dict[str, pydantic.networks.AnyUrl]] = None,
    ) -> List[Site]:
        """
        Updates all items selected by this query with the given values.

        name:                      Unique Site name
        path:                      Absolute filesystem path of the Site
        globus_endpoint_id:        Associated Globus endpoint ID
        backfill_windows:          Idle backfill currently available at the Site, keyed by queue name
        queued_jobs:               Queued scheduler jobs at the Site, keyed by scheduler ID
        optional_batch_job_params: Optional pass-through parameters accepted by the Site batchjob template
        allowed_projects:          Allowed projects/allocations for batchjob submission
        allowed_queues:            Allowed queues and associated queueing policies
        transfer_locations:        Trusted transfer location aliases and associated protocol/URLs
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._update(**kwargs)


class SiteManager(balsam._api.bases.SiteManagerBase):
    _api_path = "sites/"
    _model_class = Site
    _query_class = SiteQuery
    _bulk_create_enabled = False
    _bulk_update_enabled = False
    _bulk_delete_enabled = False
    _paginated_list_response = True

    def create(
        self,
        name: str,
        path: pathlib.Path,
        globus_endpoint_id: Optional[uuid.UUID] = None,
        backfill_windows: Optional[
            typing.Dict[str, typing.List[balsam.schemas.batchjob.SchedulerBackfillWindow]]
        ] = None,
        queued_jobs: Optional[typing.Dict[int, balsam.schemas.batchjob.SchedulerJobStatus]] = None,
        optional_batch_job_params: Optional[typing.Dict[str, str]] = None,
        allowed_projects: Optional[typing.List[str]] = None,
        allowed_queues: Optional[typing.Dict[str, balsam.schemas.site.AllowedQueue]] = None,
        transfer_locations: Optional[typing.Dict[str, pydantic.networks.AnyUrl]] = None,
    ) -> Site:
        """
        Create a new Site object and save it to the API in one step.

        name:                      Unique Site name
        path:                      Absolute filesystem path of the Site
        globus_endpoint_id:        Associated Globus endpoint ID
        backfill_windows:          Idle backfill currently available at the Site, keyed by queue name
        queued_jobs:               Queued scheduler jobs at the Site, keyed by scheduler ID
        optional_batch_job_params: Optional pass-through parameters accepted by the Site batchjob template
        allowed_projects:          Allowed projects/allocations for batchjob submission
        allowed_queues:            Allowed queues and associated queueing policies
        transfer_locations:        Trusted transfer location aliases and associated protocol/URLs
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return super()._create(**kwargs)

    def all(self) -> "SiteQuery":
        """
        Returns a Query for all Site items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        id: Union[typing.List[int], int, None] = None,
        last_refresh_after: Optional[datetime.datetime] = None,
    ) -> Site:
        """
        Retrieve exactly one Site. Raises Site.DoesNotExist
        if no items were found, or Site.MultipleObjectsReturned if
        more than one item matched the query.

        name:               Fetch the site with this name
        path:               Only return Sites with paths containing this string.
        id:                 Only return Sites having an id in this list.
        last_refresh_after: Only return Sites active since this time (UTC)
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return SiteQuery(manager=self).get(**kwargs)

    def filter(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        id: Union[typing.List[int], int, None] = None,
        last_refresh_after: Optional[datetime.datetime] = None,
    ) -> "SiteQuery":
        """
        Returns a Site Query returning items matching the filter criteria.

        name:               Fetch the site with this name
        path:               Only return Sites with paths containing this string.
        id:                 Only return Sites having an id in this list.
        last_refresh_after: Only return Sites active since this time (UTC)
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return SiteQuery(manager=self).filter(**kwargs)


class App(balsam._api.bases.AppBase):
    _create_model_cls = balsam.schemas.apps.AppCreate
    _update_model_cls = balsam.schemas.apps.AppUpdate
    _read_model_cls = balsam.schemas.apps.AppOut
    objects: "AppManager"

    site_id = Field[int]()
    name = Field[str]()
    serialized_class = Field[str]()
    source_code = Field[str]()
    description = Field[str]()
    parameters = Field[typing.Dict[str, balsam.schemas.apps.AppParameter]]()
    transfers = Field[typing.Dict[str, balsam.schemas.apps.TransferSlot]]()
    id = Field[Optional[int]]()

    def __init__(
        self,
        site_id: int,
        name: str,
        serialized_class: str,
        source_code: str,
        description: str = "",
        parameters: Optional[typing.Dict[str, balsam.schemas.apps.AppParameter]] = None,
        transfers: Optional[typing.Dict[str, balsam.schemas.apps.TransferSlot]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new App object.  You must eventually call the save() method or
        pass a App list into App.objects.bulk_create().

        site_id:          Site id at which this App is registered
        name:             Python AppDef class name
        serialized_class: Base64-encoded ApplicationDefinition payload
        source_code:      App-introspected source code
        description:      The App class docstring
        parameters:       Allowed parameters in the App command
        transfers:        Allowed transfer slots in the App
        """
        _kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        _kwargs.update(kwargs)
        return super().__init__(**_kwargs)


class AppQuery(Query[App]):
    def get(
        self,
        site_id: Union[typing.List[int], int, None] = None,
        id: Union[typing.List[int], int, None] = None,
        name: Optional[str] = None,
        site_path: Optional[str] = None,
        site_name: Optional[str] = None,
    ) -> App:
        """
        Retrieve exactly one App. Raises App.DoesNotExist
        if no items were found, or App.MultipleObjectsReturned if
        more than one item matched the query.

        site_id:   Only return Apps associated with the Site IDs in this list.
        id:        Only return Apps with IDs in this list.
        name:      Only return Apps having an ApplicationDefinition with this class name.
        site_path: Only return Apps from Sites having paths containing this substring.
        site_name: Only return Apps from the Site with this unique name.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        site_id: Union[typing.List[int], int, None] = None,
        id: Union[typing.List[int], int, None] = None,
        name: Optional[str] = None,
        site_path: Optional[str] = None,
        site_name: Optional[str] = None,
    ) -> "AppQuery":
        """
        Retrieve exactly one App. Raises App.DoesNotExist
        if no items were found, or App.MultipleObjectsReturned if
        more than one item matched the query.

        site_id:   Only return Apps associated with the Site IDs in this list.
        id:        Only return Apps with IDs in this list.
        name:      Only return Apps having an ApplicationDefinition with this class name.
        site_path: Only return Apps from Sites having paths containing this substring.
        site_name: Only return Apps from the Site with this unique name.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)

    def update(
        self,
        site_id: Optional[int] = None,
        name: Optional[str] = None,
        serialized_class: Optional[str] = None,
        source_code: Optional[str] = None,
        description: Optional[str] = None,
        parameters: Optional[typing.Dict[str, balsam.schemas.apps.AppParameter]] = None,
        transfers: Optional[typing.Dict[str, balsam.schemas.apps.TransferSlot]] = None,
    ) -> List[App]:
        """
        Updates all items selected by this query with the given values.

        site_id:          Site id at which this App is registered
        name:             Python AppDef class name
        serialized_class: Base64-encoded ApplicationDefinition payload
        source_code:      App-introspected source code
        description:      The App class docstring
        parameters:       Allowed parameters in the App command
        transfers:        Allowed transfer slots in the App
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._update(**kwargs)


class AppManager(balsam._api.bases.AppManagerBase):
    _api_path = "apps/"
    _model_class = App
    _query_class = AppQuery
    _bulk_create_enabled = False
    _bulk_update_enabled = False
    _bulk_delete_enabled = False
    _paginated_list_response = True

    def create(
        self,
        site_id: int,
        name: str,
        serialized_class: str,
        source_code: str,
        description: str = "",
        parameters: Optional[typing.Dict[str, balsam.schemas.apps.AppParameter]] = None,
        transfers: Optional[typing.Dict[str, balsam.schemas.apps.TransferSlot]] = None,
    ) -> App:
        """
        Create a new App object and save it to the API in one step.

        site_id:          Site id at which this App is registered
        name:             Python AppDef class name
        serialized_class: Base64-encoded ApplicationDefinition payload
        source_code:      App-introspected source code
        description:      The App class docstring
        parameters:       Allowed parameters in the App command
        transfers:        Allowed transfer slots in the App
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return super()._create(**kwargs)

    def all(self) -> "AppQuery":
        """
        Returns a Query for all App items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        site_id: Union[typing.List[int], int, None] = None,
        id: Union[typing.List[int], int, None] = None,
        name: Optional[str] = None,
        site_path: Optional[str] = None,
        site_name: Optional[str] = None,
    ) -> App:
        """
        Retrieve exactly one App. Raises App.DoesNotExist
        if no items were found, or App.MultipleObjectsReturned if
        more than one item matched the query.

        site_id:   Only return Apps associated with the Site IDs in this list.
        id:        Only return Apps with IDs in this list.
        name:      Only return Apps having an ApplicationDefinition with this class name.
        site_path: Only return Apps from Sites having paths containing this substring.
        site_name: Only return Apps from the Site with this unique name.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return AppQuery(manager=self).get(**kwargs)

    def filter(
        self,
        site_id: Union[typing.List[int], int, None] = None,
        id: Union[typing.List[int], int, None] = None,
        name: Optional[str] = None,
        site_path: Optional[str] = None,
        site_name: Optional[str] = None,
    ) -> "AppQuery":
        """
        Returns a App Query returning items matching the filter criteria.

        site_id:   Only return Apps associated with the Site IDs in this list.
        id:        Only return Apps with IDs in this list.
        name:      Only return Apps having an ApplicationDefinition with this class name.
        site_path: Only return Apps from Sites having paths containing this substring.
        site_name: Only return Apps from the Site with this unique name.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return AppQuery(manager=self).filter(**kwargs)


class Job(balsam._api.bases.JobBase):
    _create_model_cls = balsam.schemas.job.JobCreate
    _update_model_cls = balsam.schemas.job.JobUpdate
    _read_model_cls = balsam.schemas.job.JobOut
    objects: "JobManager"

    workdir = Field[pathlib.Path]()
    tags = Field[typing.Dict[str, str]]()
    data = Field[typing.Dict[str, typing.Any]]()
    return_code = Field[Optional[int]]()
    num_nodes = Field[int]()
    ranks_per_node = Field[int]()
    threads_per_rank = Field[int]()
    threads_per_core = Field[int]()
    launch_params = Field[typing.Dict[str, str]]()
    gpus_per_rank = Field[float]()
    node_packing_count = Field[int]()
    wall_time_min = Field[int]()
    app_id = Field[int]()
    parent_ids = Field[typing.Set[int]]()
    transfers = Field[typing.Dict[str, balsam.schemas.job.JobTransferItem]]()
    batch_job_id = Field[Optional[int]]()
    state = Field[Optional[balsam.schemas.job.JobState]]()
    state_timestamp = Field[Optional[datetime.datetime]]()
    state_data = Field[Optional[typing.Dict[str, typing.Any]]]()
    pending_file_cleanup = Field[Optional[bool]]()
    id = Field[Optional[int]]()
    last_update = Field[Optional[datetime.datetime]]()

    def __init__(
        self,
        workdir: pathlib.Path,
        app_id: typing.Union[int, str],
        tags: Optional[typing.Dict[str, str]] = None,
        data: Optional[typing.Dict[str, typing.Any]] = None,
        return_code: Optional[int] = None,
        num_nodes: int = 1,
        ranks_per_node: int = 1,
        threads_per_rank: int = 1,
        threads_per_core: int = 1,
        launch_params: Optional[typing.Dict[str, str]] = None,
        gpus_per_rank: float = 0,
        node_packing_count: int = 1,
        wall_time_min: int = 0,
        site_name: Optional[str] = None,
        parameters: Optional[typing.Dict[str, typing.Any]] = None,
        parent_ids: typing.Set[int] = set(),
        transfers: Optional[typing.Dict[str, balsam.schemas.job.JobTransferItem]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new Job object.  You must eventually call the save() method or
        pass a Job list into Job.objects.bulk_create().

        workdir:            Job path relative to site data/ folder.
        tags:               Custom key:value string tags.
        data:               Arbitrary JSON-able data dictionary.
        return_code:        Return code from last execution of this Job.
        num_nodes:          Number of compute nodes needed.
        ranks_per_node:     Number of MPI processes per node.
        threads_per_rank:   Logical threads per process.
        threads_per_core:   Logical threads per CPU core.
        launch_params:      Optional pass-through parameters to MPI application launcher.
        gpus_per_rank:      Number of GPUs per process.
        node_packing_count: Maximum number of concurrent runs per node.
        wall_time_min:      Optional estimate of Job runtime. All else being equal, longer Jobs tend to run first.
        app_id:             App name, ID, or class
        site_name:          Site name, to disambiguate app defined at multiple Sites.
        parameters:         Parameters passed to App at runtime.
        parent_ids:         Set of parent Job IDs (dependencies).
        transfers:          TransferItem dictionary. One key:JobTransferItem pair for each slot defined on the App.
        """
        _kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        _kwargs.update(kwargs)
        return super().__init__(**_kwargs)


class JobQuery(Query[Job]):
    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        parent_id: Union[typing.List[int], int, None] = None,
        app_id: Optional[int] = None,
        site_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        last_update_before: Optional[datetime.datetime] = None,
        last_update_after: Optional[datetime.datetime] = None,
        workdir__contains: Optional[str] = None,
        state__ne: Optional[balsam.schemas.job.JobState] = None,
        state: Union[typing.Set[balsam.schemas.job.JobState], balsam.schemas.job.JobState, None] = None,
        tags: Union[typing.List[str], str, None] = None,
        pending_file_cleanup: Optional[bool] = None,
    ) -> Job:
        """
        Retrieve exactly one Job. Raises Job.DoesNotExist
        if no items were found, or Job.MultipleObjectsReturned if
        more than one item matched the query.

        id:                   Only return Jobs with ids in this list.
        parent_id:            Only return Jobs that are children of Jobs with ids in this list.
        app_id:               Only return Jobs associated with this App id.
        site_id:              Only return Jobs associated with these Site ids.
        batch_job_id:         Only return Jobs associated with this BatchJob id.
        last_update_before:   Only return Jobs that were updated before this time (UTC).
        last_update_after:    Only return Jobs that were updated after this time (UTC).
        workdir__contains:    Only return jobs with workdirs containing this string.
        state__ne:            Only return jobs with states not equal to this state.
        state:                Only return jobs in this set of states.
        tags:                 Only return jobs containing these tags (list of KEY:VALUE strings)
        pending_file_cleanup: Only return jobs which have not yet had workdir cleaned.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        parent_id: Union[typing.List[int], int, None] = None,
        app_id: Optional[int] = None,
        site_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        last_update_before: Optional[datetime.datetime] = None,
        last_update_after: Optional[datetime.datetime] = None,
        workdir__contains: Optional[str] = None,
        state__ne: Optional[balsam.schemas.job.JobState] = None,
        state: Union[typing.Set[balsam.schemas.job.JobState], balsam.schemas.job.JobState, None] = None,
        tags: Union[typing.List[str], str, None] = None,
        pending_file_cleanup: Optional[bool] = None,
    ) -> "JobQuery":
        """
        Retrieve exactly one Job. Raises Job.DoesNotExist
        if no items were found, or Job.MultipleObjectsReturned if
        more than one item matched the query.

        id:                   Only return Jobs with ids in this list.
        parent_id:            Only return Jobs that are children of Jobs with ids in this list.
        app_id:               Only return Jobs associated with this App id.
        site_id:              Only return Jobs associated with these Site ids.
        batch_job_id:         Only return Jobs associated with this BatchJob id.
        last_update_before:   Only return Jobs that were updated before this time (UTC).
        last_update_after:    Only return Jobs that were updated after this time (UTC).
        workdir__contains:    Only return jobs with workdirs containing this string.
        state__ne:            Only return jobs with states not equal to this state.
        state:                Only return jobs in this set of states.
        tags:                 Only return jobs containing these tags (list of KEY:VALUE strings)
        pending_file_cleanup: Only return jobs which have not yet had workdir cleaned.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)

    def update(
        self,
        workdir: Optional[pathlib.Path] = None,
        tags: Optional[typing.Dict[str, str]] = None,
        serialized_parameters: Optional[str] = None,
        data: Optional[typing.Dict[str, typing.Any]] = None,
        return_code: Optional[int] = None,
        num_nodes: Optional[int] = None,
        ranks_per_node: Optional[int] = None,
        threads_per_rank: Optional[int] = None,
        threads_per_core: Optional[int] = None,
        launch_params: Optional[typing.Dict[str, str]] = None,
        gpus_per_rank: Optional[float] = None,
        node_packing_count: Optional[int] = None,
        wall_time_min: Optional[int] = None,
        batch_job_id: Optional[int] = None,
        state: Optional[balsam.schemas.job.JobState] = None,
        state_timestamp: Optional[datetime.datetime] = None,
        state_data: Optional[typing.Dict[str, typing.Any]] = None,
        pending_file_cleanup: Optional[bool] = None,
        serialized_return_value: Optional[str] = None,
        serialized_exception: Optional[str] = None,
    ) -> List[Job]:
        """
        Updates all items selected by this query with the given values.

        workdir:                 Job path relative to the site data/ folder
        tags:                    Custom key:value string tags.
        serialized_parameters:   Encoded parameters dict
        data:                    Arbitrary JSON-able data dictionary.
        return_code:             Return code from last execution of this Job.
        num_nodes:               Number of compute nodes needed.
        ranks_per_node:          Number of MPI processes per node.
        threads_per_rank:        Logical threads per process.
        threads_per_core:        Logical threads per CPU core.
        launch_params:           Optional pass-through parameters to MPI application launcher.
        gpus_per_rank:           Number of GPUs per process.
        node_packing_count:      Maximum number of concurrent runs per node.
        wall_time_min:           Optional estimate of Job runtime. All else being equal, longer Jobs tend to run first.
        batch_job_id:            ID of most recent BatchJob in which this Job ran
        state:                   Job state
        state_timestamp:         Time (UTC) at which Job state change occured
        state_data:              Arbitrary associated state change data for logging
        pending_file_cleanup:    Whether job remains to have workdir cleaned.
        serialized_return_value: Encoded return value
        serialized_exception:    Encoded wrapped Exception
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._update(**kwargs)

    def order_by(self, field: Optional[balsam.schemas.job.JobOrdering]) -> "JobQuery":
        """
        Order the returned items by this field.
        """
        return self._order_by(field)


class JobManager(balsam._api.bases.JobManagerBase):
    _api_path = "jobs/"
    _model_class = Job
    _query_class = JobQuery
    _bulk_create_enabled = True
    _bulk_update_enabled = True
    _bulk_delete_enabled = True
    _paginated_list_response = True

    def create(
        self,
        workdir: pathlib.Path,
        app_id: typing.Union[int, str],
        tags: Optional[typing.Dict[str, str]] = None,
        data: Optional[typing.Dict[str, typing.Any]] = None,
        return_code: Optional[int] = None,
        num_nodes: int = 1,
        ranks_per_node: int = 1,
        threads_per_rank: int = 1,
        threads_per_core: int = 1,
        launch_params: Optional[typing.Dict[str, str]] = None,
        gpus_per_rank: float = 0,
        node_packing_count: int = 1,
        wall_time_min: int = 0,
        site_name: Optional[str] = None,
        parameters: Optional[typing.Dict[str, typing.Any]] = None,
        parent_ids: typing.Set[int] = set(),
        transfers: Optional[typing.Dict[str, balsam.schemas.job.JobTransferItem]] = None,
    ) -> Job:
        """
        Create a new Job object and save it to the API in one step.

        workdir:            Job path relative to site data/ folder.
        tags:               Custom key:value string tags.
        data:               Arbitrary JSON-able data dictionary.
        return_code:        Return code from last execution of this Job.
        num_nodes:          Number of compute nodes needed.
        ranks_per_node:     Number of MPI processes per node.
        threads_per_rank:   Logical threads per process.
        threads_per_core:   Logical threads per CPU core.
        launch_params:      Optional pass-through parameters to MPI application launcher.
        gpus_per_rank:      Number of GPUs per process.
        node_packing_count: Maximum number of concurrent runs per node.
        wall_time_min:      Optional estimate of Job runtime. All else being equal, longer Jobs tend to run first.
        app_id:             App name, ID, or class
        site_name:          Site name, to disambiguate app defined at multiple Sites.
        parameters:         Parameters passed to App at runtime.
        parent_ids:         Set of parent Job IDs (dependencies).
        transfers:          TransferItem dictionary. One key:JobTransferItem pair for each slot defined on the App.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return super()._create(**kwargs)

    def all(self) -> "JobQuery":
        """
        Returns a Query for all Job items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        parent_id: Union[typing.List[int], int, None] = None,
        app_id: Optional[int] = None,
        site_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        last_update_before: Optional[datetime.datetime] = None,
        last_update_after: Optional[datetime.datetime] = None,
        workdir__contains: Optional[str] = None,
        state__ne: Optional[balsam.schemas.job.JobState] = None,
        state: Union[typing.Set[balsam.schemas.job.JobState], balsam.schemas.job.JobState, None] = None,
        tags: Union[typing.List[str], str, None] = None,
        pending_file_cleanup: Optional[bool] = None,
    ) -> Job:
        """
        Retrieve exactly one Job. Raises Job.DoesNotExist
        if no items were found, or Job.MultipleObjectsReturned if
        more than one item matched the query.

        id:                   Only return Jobs with ids in this list.
        parent_id:            Only return Jobs that are children of Jobs with ids in this list.
        app_id:               Only return Jobs associated with this App id.
        site_id:              Only return Jobs associated with these Site ids.
        batch_job_id:         Only return Jobs associated with this BatchJob id.
        last_update_before:   Only return Jobs that were updated before this time (UTC).
        last_update_after:    Only return Jobs that were updated after this time (UTC).
        workdir__contains:    Only return jobs with workdirs containing this string.
        state__ne:            Only return jobs with states not equal to this state.
        state:                Only return jobs in this set of states.
        tags:                 Only return jobs containing these tags (list of KEY:VALUE strings)
        pending_file_cleanup: Only return jobs which have not yet had workdir cleaned.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return JobQuery(manager=self).get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        parent_id: Union[typing.List[int], int, None] = None,
        app_id: Optional[int] = None,
        site_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        last_update_before: Optional[datetime.datetime] = None,
        last_update_after: Optional[datetime.datetime] = None,
        workdir__contains: Optional[str] = None,
        state__ne: Optional[balsam.schemas.job.JobState] = None,
        state: Union[typing.Set[balsam.schemas.job.JobState], balsam.schemas.job.JobState, None] = None,
        tags: Union[typing.List[str], str, None] = None,
        pending_file_cleanup: Optional[bool] = None,
    ) -> "JobQuery":
        """
        Returns a Job Query returning items matching the filter criteria.

        id:                   Only return Jobs with ids in this list.
        parent_id:            Only return Jobs that are children of Jobs with ids in this list.
        app_id:               Only return Jobs associated with this App id.
        site_id:              Only return Jobs associated with these Site ids.
        batch_job_id:         Only return Jobs associated with this BatchJob id.
        last_update_before:   Only return Jobs that were updated before this time (UTC).
        last_update_after:    Only return Jobs that were updated after this time (UTC).
        workdir__contains:    Only return jobs with workdirs containing this string.
        state__ne:            Only return jobs with states not equal to this state.
        state:                Only return jobs in this set of states.
        tags:                 Only return jobs containing these tags (list of KEY:VALUE strings)
        pending_file_cleanup: Only return jobs which have not yet had workdir cleaned.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return JobQuery(manager=self).filter(**kwargs)


class BatchJob(balsam._api.bases.BatchJobBase):
    _create_model_cls = balsam.schemas.batchjob.BatchJobCreate
    _update_model_cls = balsam.schemas.batchjob.BatchJobUpdate
    _read_model_cls = balsam.schemas.batchjob.BatchJobOut
    objects: "BatchJobManager"

    num_nodes = Field[int]()
    wall_time_min = Field[int]()
    job_mode = Field[balsam.schemas.batchjob.JobMode]()
    optional_params = Field[typing.Dict[str, str]]()
    filter_tags = Field[typing.Dict[str, str]]()
    partitions = Field[typing.Optional[typing.List[balsam.schemas.batchjob.BatchJobPartition]]]()
    site_id = Field[int]()
    project = Field[str]()
    queue = Field[str]()
    scheduler_id = Field[Optional[int]]()
    state = Field[Optional[balsam.schemas.batchjob.BatchJobState]]()
    status_info = Field[Optional[typing.Dict[str, str]]]()
    start_time = Field[Optional[datetime.datetime]]()
    end_time = Field[Optional[datetime.datetime]]()
    id = Field[Optional[int]]()

    def __init__(
        self,
        num_nodes: int,
        wall_time_min: int,
        job_mode: balsam.schemas.batchjob.JobMode,
        site_id: int,
        project: str,
        queue: str,
        optional_params: Optional[typing.Dict[str, str]] = None,
        filter_tags: Optional[typing.Dict[str, str]] = None,
        partitions: Optional[typing.Optional[typing.List[balsam.schemas.batchjob.BatchJobPartition]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new BatchJob object.  You must eventually call the save() method or
        pass a BatchJob list into BatchJob.objects.bulk_create().

        num_nodes:       Requested number of nodes for this allocation
        wall_time_min:   Requested wall clock time for this allocation
        job_mode:        Balsam launcher execution mode (if single partition)
        optional_params: Optional pass-through parameters submitted with the batchjob script
        filter_tags:     Only run Jobs containing these tags
        partitions:      Optionally, subdivide an allocation into multiple partitions.
        site_id:         The Site id where this batchjob is submitted
        project:         The project/allocation to charge for this batchjob
        queue:           Which queue the batchjob is submitted on
        """
        _kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        _kwargs.update(kwargs)
        return super().__init__(**_kwargs)


class BatchJobQuery(Query[BatchJob]):
    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Union[typing.List[int], int, None] = None,
        state: Union[typing.List[str], str, None] = None,
        scheduler_id: Optional[int] = None,
        queue: Optional[str] = None,
        start_time_before: Optional[datetime.datetime] = None,
        start_time_after: Optional[datetime.datetime] = None,
        end_time_before: Optional[datetime.datetime] = None,
        end_time_after: Optional[datetime.datetime] = None,
        filter_tags: Union[typing.List[str], str, None] = None,
    ) -> BatchJob:
        """
        Retrieve exactly one BatchJob. Raises BatchJob.DoesNotExist
        if no items were found, or BatchJob.MultipleObjectsReturned if
        more than one item matched the query.

        id:                Only return BatchJobs having an id in this list.
        site_id:           Only return batchjobs for Sites in this id list.
        state:             Only return batchjobs having one of these States in this list.
        scheduler_id:      Return the batchjob with this local scheduler id.
        queue:             Only return batchjobs submitted to this queue.
        start_time_before: Only return batchjobs that started before this time (UTC).
        start_time_after:  Only return batchjobs that started after this time (UTC).
        end_time_before:   Only return batchjobs that finished before this time (UTC).
        end_time_after:    Only return batchjobs that finished after this time (UTC).
        filter_tags:       Only return batchjobs processing these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Union[typing.List[int], int, None] = None,
        state: Union[typing.List[str], str, None] = None,
        scheduler_id: Optional[int] = None,
        queue: Optional[str] = None,
        start_time_before: Optional[datetime.datetime] = None,
        start_time_after: Optional[datetime.datetime] = None,
        end_time_before: Optional[datetime.datetime] = None,
        end_time_after: Optional[datetime.datetime] = None,
        filter_tags: Union[typing.List[str], str, None] = None,
    ) -> "BatchJobQuery":
        """
        Retrieve exactly one BatchJob. Raises BatchJob.DoesNotExist
        if no items were found, or BatchJob.MultipleObjectsReturned if
        more than one item matched the query.

        id:                Only return BatchJobs having an id in this list.
        site_id:           Only return batchjobs for Sites in this id list.
        state:             Only return batchjobs having one of these States in this list.
        scheduler_id:      Return the batchjob with this local scheduler id.
        queue:             Only return batchjobs submitted to this queue.
        start_time_before: Only return batchjobs that started before this time (UTC).
        start_time_after:  Only return batchjobs that started after this time (UTC).
        end_time_before:   Only return batchjobs that finished before this time (UTC).
        end_time_after:    Only return batchjobs that finished after this time (UTC).
        filter_tags:       Only return batchjobs processing these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)

    def update(
        self,
        scheduler_id: Optional[int] = None,
        state: Optional[balsam.schemas.batchjob.BatchJobState] = None,
        status_info: Optional[typing.Dict[str, str]] = None,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
    ) -> List[BatchJob]:
        """
        Updates all items selected by this query with the given values.

        scheduler_id: The local HPC scheduler's ID for this batchjob
        state:        Status of this batchjob in the local HPC scheduler
        status_info:  Arbitrary status info
        start_time:   BatchJob execution start time
        end_time:     BatchJob execution end time
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._update(**kwargs)

    def order_by(self, field: Optional[balsam.schemas.batchjob.BatchJobOrdering]) -> "BatchJobQuery":
        """
        Order the returned items by this field.
        """
        return self._order_by(field)


class BatchJobManager(balsam._api.bases.BatchJobManagerBase):
    _api_path = "batch-jobs/"
    _model_class = BatchJob
    _query_class = BatchJobQuery
    _bulk_create_enabled = False
    _bulk_update_enabled = True
    _bulk_delete_enabled = False
    _paginated_list_response = True

    def create(
        self,
        num_nodes: int,
        wall_time_min: int,
        job_mode: balsam.schemas.batchjob.JobMode,
        site_id: int,
        project: str,
        queue: str,
        optional_params: Optional[typing.Dict[str, str]] = None,
        filter_tags: Optional[typing.Dict[str, str]] = None,
        partitions: Optional[typing.Optional[typing.List[balsam.schemas.batchjob.BatchJobPartition]]] = None,
    ) -> BatchJob:
        """
        Create a new BatchJob object and save it to the API in one step.

        num_nodes:       Requested number of nodes for this allocation
        wall_time_min:   Requested wall clock time for this allocation
        job_mode:        Balsam launcher execution mode (if single partition)
        optional_params: Optional pass-through parameters submitted with the batchjob script
        filter_tags:     Only run Jobs containing these tags
        partitions:      Optionally, subdivide an allocation into multiple partitions.
        site_id:         The Site id where this batchjob is submitted
        project:         The project/allocation to charge for this batchjob
        queue:           Which queue the batchjob is submitted on
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return super()._create(**kwargs)

    def all(self) -> "BatchJobQuery":
        """
        Returns a Query for all BatchJob items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Union[typing.List[int], int, None] = None,
        state: Union[typing.List[str], str, None] = None,
        scheduler_id: Optional[int] = None,
        queue: Optional[str] = None,
        start_time_before: Optional[datetime.datetime] = None,
        start_time_after: Optional[datetime.datetime] = None,
        end_time_before: Optional[datetime.datetime] = None,
        end_time_after: Optional[datetime.datetime] = None,
        filter_tags: Union[typing.List[str], str, None] = None,
    ) -> BatchJob:
        """
        Retrieve exactly one BatchJob. Raises BatchJob.DoesNotExist
        if no items were found, or BatchJob.MultipleObjectsReturned if
        more than one item matched the query.

        id:                Only return BatchJobs having an id in this list.
        site_id:           Only return batchjobs for Sites in this id list.
        state:             Only return batchjobs having one of these States in this list.
        scheduler_id:      Return the batchjob with this local scheduler id.
        queue:             Only return batchjobs submitted to this queue.
        start_time_before: Only return batchjobs that started before this time (UTC).
        start_time_after:  Only return batchjobs that started after this time (UTC).
        end_time_before:   Only return batchjobs that finished before this time (UTC).
        end_time_after:    Only return batchjobs that finished after this time (UTC).
        filter_tags:       Only return batchjobs processing these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return BatchJobQuery(manager=self).get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Union[typing.List[int], int, None] = None,
        state: Union[typing.List[str], str, None] = None,
        scheduler_id: Optional[int] = None,
        queue: Optional[str] = None,
        start_time_before: Optional[datetime.datetime] = None,
        start_time_after: Optional[datetime.datetime] = None,
        end_time_before: Optional[datetime.datetime] = None,
        end_time_after: Optional[datetime.datetime] = None,
        filter_tags: Union[typing.List[str], str, None] = None,
    ) -> "BatchJobQuery":
        """
        Returns a BatchJob Query returning items matching the filter criteria.

        id:                Only return BatchJobs having an id in this list.
        site_id:           Only return batchjobs for Sites in this id list.
        state:             Only return batchjobs having one of these States in this list.
        scheduler_id:      Return the batchjob with this local scheduler id.
        queue:             Only return batchjobs submitted to this queue.
        start_time_before: Only return batchjobs that started before this time (UTC).
        start_time_after:  Only return batchjobs that started after this time (UTC).
        end_time_before:   Only return batchjobs that finished before this time (UTC).
        end_time_after:    Only return batchjobs that finished after this time (UTC).
        filter_tags:       Only return batchjobs processing these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return BatchJobQuery(manager=self).filter(**kwargs)


class Session(balsam._api.bases.SessionBase):
    _create_model_cls = balsam.schemas.session.SessionCreate
    _update_model_cls = None
    _read_model_cls = balsam.schemas.session.SessionOut
    objects: "SessionManager"

    site_id = Field[int]()
    batch_job_id = Field[Optional[int]]()
    id = Field[Optional[int]]()
    heartbeat = Field[Optional[datetime.datetime]]()

    def __init__(
        self,
        site_id: int,
        batch_job_id: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """
        Construct a new Session object.  You must eventually call the save() method or
        pass a Session list into Session.objects.bulk_create().

        site_id:      Site id of the running Session
        batch_job_id: Associated batchjob id
        """
        _kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        _kwargs.update(kwargs)
        return super().__init__(**_kwargs)


class SessionQuery(Query[Session]):
    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
    ) -> Session:
        """
        Retrieve exactly one Session. Raises Session.DoesNotExist
        if no items were found, or Session.MultipleObjectsReturned if
        more than one item matched the query.

        id: Only return Sessions having an id in this list.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
    ) -> "SessionQuery":
        """
        Retrieve exactly one Session. Raises Session.DoesNotExist
        if no items were found, or Session.MultipleObjectsReturned if
        more than one item matched the query.

        id: Only return Sessions having an id in this list.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)


class SessionManager(balsam._api.bases.SessionManagerBase):
    _api_path = "sessions/"
    _model_class = Session
    _query_class = SessionQuery
    _bulk_create_enabled = False
    _bulk_update_enabled = False
    _bulk_delete_enabled = False
    _paginated_list_response = True

    def create(
        self,
        site_id: int,
        batch_job_id: Optional[int] = None,
    ) -> Session:
        """
        Create a new Session object and save it to the API in one step.

        site_id:      Site id of the running Session
        batch_job_id: Associated batchjob id
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return super()._create(**kwargs)

    def all(self) -> "SessionQuery":
        """
        Returns a Query for all Session items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
    ) -> Session:
        """
        Retrieve exactly one Session. Raises Session.DoesNotExist
        if no items were found, or Session.MultipleObjectsReturned if
        more than one item matched the query.

        id: Only return Sessions having an id in this list.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return SessionQuery(manager=self).get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
    ) -> "SessionQuery":
        """
        Returns a Session Query returning items matching the filter criteria.

        id: Only return Sessions having an id in this list.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return SessionQuery(manager=self).filter(**kwargs)


class TransferItem(balsam._api.bases.TransferItemBase):
    _create_model_cls = None
    _update_model_cls = balsam.schemas.transfer.TransferItemUpdate
    _read_model_cls = balsam.schemas.transfer.TransferItemOut
    objects: "TransferItemManager"

    state = Field[balsam.schemas.transfer.TransferItemState]()
    task_id = Field[str]()
    transfer_info = Field[typing.Dict[str, typing.Any]]()
    id = Field[int]()
    job_id = Field[int]()
    direction = Field[balsam.schemas.transfer.TransferDirection]()
    local_path = Field[pathlib.Path]()
    remote_path = Field[pathlib.Path]()
    location_alias = Field[str]()
    recursive = Field[bool]()


class TransferItemQuery(Query[TransferItem]):
    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Optional[int] = None,
        job_id: Union[typing.List[int], int, None] = None,
        state: Union[
            typing.Set[balsam.schemas.transfer.TransferItemState], balsam.schemas.transfer.TransferItemState, None
        ] = None,
        direction: Optional[balsam.schemas.transfer.TransferDirection] = None,
        job_state: Optional[str] = None,
        tags: Union[typing.List[str], str, None] = None,
    ) -> TransferItem:
        """
        Retrieve exactly one TransferItem. Raises TransferItem.DoesNotExist
        if no items were found, or TransferItem.MultipleObjectsReturned if
        more than one item matched the query.

        id:        Only return transfer items with IDs in this list.
        site_id:   Only return transfer items associated with this Site id.
        job_id:    Only return transfer items associated with this Job id list.
        state:     Only return transfer items in this set of states.
        direction: Only return items in this transfer direction.
        job_state: Only return transfer items for Jobs having this state.
        tags:      Only return transfer items for Jobs having these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Optional[int] = None,
        job_id: Union[typing.List[int], int, None] = None,
        state: Union[
            typing.Set[balsam.schemas.transfer.TransferItemState], balsam.schemas.transfer.TransferItemState, None
        ] = None,
        direction: Optional[balsam.schemas.transfer.TransferDirection] = None,
        job_state: Optional[str] = None,
        tags: Union[typing.List[str], str, None] = None,
    ) -> "TransferItemQuery":
        """
        Retrieve exactly one TransferItem. Raises TransferItem.DoesNotExist
        if no items were found, or TransferItem.MultipleObjectsReturned if
        more than one item matched the query.

        id:        Only return transfer items with IDs in this list.
        site_id:   Only return transfer items associated with this Site id.
        job_id:    Only return transfer items associated with this Job id list.
        state:     Only return transfer items in this set of states.
        direction: Only return items in this transfer direction.
        job_state: Only return transfer items for Jobs having this state.
        tags:      Only return transfer items for Jobs having these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)

    def update(
        self,
        state: Optional[balsam.schemas.transfer.TransferItemState] = None,
        task_id: Optional[str] = None,
        transfer_info: Optional[typing.Dict[str, typing.Any]] = None,
    ) -> List[TransferItem]:
        """
        Updates all items selected by this query with the given values.

        state:         Status of this transfer item
        task_id:       Transfer Task ID used to lookup transfer item status
        transfer_info: Arbitrary transfer state info
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._update(**kwargs)


class TransferItemManager(balsam._api.bases.TransferItemManagerBase):
    _api_path = "transfers/"
    _model_class = TransferItem
    _query_class = TransferItemQuery
    _bulk_create_enabled = False
    _bulk_update_enabled = True
    _bulk_delete_enabled = False
    _paginated_list_response = True

    def all(self) -> "TransferItemQuery":
        """
        Returns a Query for all TransferItem items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Optional[int] = None,
        job_id: Union[typing.List[int], int, None] = None,
        state: Union[
            typing.Set[balsam.schemas.transfer.TransferItemState], balsam.schemas.transfer.TransferItemState, None
        ] = None,
        direction: Optional[balsam.schemas.transfer.TransferDirection] = None,
        job_state: Optional[str] = None,
        tags: Union[typing.List[str], str, None] = None,
    ) -> TransferItem:
        """
        Retrieve exactly one TransferItem. Raises TransferItem.DoesNotExist
        if no items were found, or TransferItem.MultipleObjectsReturned if
        more than one item matched the query.

        id:        Only return transfer items with IDs in this list.
        site_id:   Only return transfer items associated with this Site id.
        job_id:    Only return transfer items associated with this Job id list.
        state:     Only return transfer items in this set of states.
        direction: Only return items in this transfer direction.
        job_state: Only return transfer items for Jobs having this state.
        tags:      Only return transfer items for Jobs having these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return TransferItemQuery(manager=self).get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        site_id: Optional[int] = None,
        job_id: Union[typing.List[int], int, None] = None,
        state: Union[
            typing.Set[balsam.schemas.transfer.TransferItemState], balsam.schemas.transfer.TransferItemState, None
        ] = None,
        direction: Optional[balsam.schemas.transfer.TransferDirection] = None,
        job_state: Optional[str] = None,
        tags: Union[typing.List[str], str, None] = None,
    ) -> "TransferItemQuery":
        """
        Returns a TransferItem Query returning items matching the filter criteria.

        id:        Only return transfer items with IDs in this list.
        site_id:   Only return transfer items associated with this Site id.
        job_id:    Only return transfer items associated with this Job id list.
        state:     Only return transfer items in this set of states.
        direction: Only return items in this transfer direction.
        job_state: Only return transfer items for Jobs having this state.
        tags:      Only return transfer items for Jobs having these tags (list of KEY:VALUE strings).
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return TransferItemQuery(manager=self).filter(**kwargs)


class EventLog(balsam._api.bases.EventLogBase):
    _create_model_cls = None
    _update_model_cls = None
    _read_model_cls = balsam.schemas.logevent.LogEventOut
    objects: "EventLogManager"

    id = Field[int]()
    job_id = Field[int]()
    timestamp = Field[datetime.datetime]()
    from_state = Field[str]()
    to_state = Field[str]()
    data = Field[typing.Dict[str, typing.Any]]()


class EventLogQuery(Query[EventLog]):
    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        job_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        tags: Union[typing.List[str], str, None] = None,
        data: Union[typing.List[str], str, None] = None,
        timestamp_before: Optional[datetime.datetime] = None,
        timestamp_after: Optional[datetime.datetime] = None,
        from_state: Optional[str] = None,
        to_state: Optional[str] = None,
    ) -> EventLog:
        """
        Retrieve exactly one EventLog. Raises EventLog.DoesNotExist
        if no items were found, or EventLog.MultipleObjectsReturned if
        more than one item matched the query.

        id:               Only return EventLogs having an id in this list.
        job_id:           Only return Events associated with Job IDs in this list.
        batch_job_id:     Only return Events associated this BatchJob id.
        scheduler_id:     Only return Events associated with this HPC scheduler job ID.
        tags:             Only return Events for Jobs containing these tags (list of KEY:VALUE strings)
        data:             Only return Events containing this data (list of KEY:VALUE strings)
        timestamp_before: Only return Events before this time (UTC).
        timestamp_after:  Only return Events that occured after this time (UTC).
        from_state:       Only return Events transitioning from this Job state.
        to_state:         Only return Events transitioning to this Job state.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        job_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        tags: Union[typing.List[str], str, None] = None,
        data: Union[typing.List[str], str, None] = None,
        timestamp_before: Optional[datetime.datetime] = None,
        timestamp_after: Optional[datetime.datetime] = None,
        from_state: Optional[str] = None,
        to_state: Optional[str] = None,
    ) -> "EventLogQuery":
        """
        Retrieve exactly one EventLog. Raises EventLog.DoesNotExist
        if no items were found, or EventLog.MultipleObjectsReturned if
        more than one item matched the query.

        id:               Only return EventLogs having an id in this list.
        job_id:           Only return Events associated with Job IDs in this list.
        batch_job_id:     Only return Events associated this BatchJob id.
        scheduler_id:     Only return Events associated with this HPC scheduler job ID.
        tags:             Only return Events for Jobs containing these tags (list of KEY:VALUE strings)
        data:             Only return Events containing this data (list of KEY:VALUE strings)
        timestamp_before: Only return Events before this time (UTC).
        timestamp_after:  Only return Events that occured after this time (UTC).
        from_state:       Only return Events transitioning from this Job state.
        to_state:         Only return Events transitioning to this Job state.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return self._filter(**kwargs)

    def order_by(self, field: Optional[balsam.schemas.logevent.EventOrdering]) -> "EventLogQuery":
        """
        Order the returned items by this field.
        """
        return self._order_by(field)


class EventLogManager(balsam._api.bases.EventLogManagerBase):
    _api_path = "events/"
    _model_class = EventLog
    _query_class = EventLogQuery
    _bulk_create_enabled = False
    _bulk_update_enabled = False
    _bulk_delete_enabled = False
    _paginated_list_response = True

    def all(self) -> "EventLogQuery":
        """
        Returns a Query for all EventLog items.
        """
        return self._query_class(manager=self)

    def get(
        self,
        id: Union[typing.List[int], int, None] = None,
        job_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        tags: Union[typing.List[str], str, None] = None,
        data: Union[typing.List[str], str, None] = None,
        timestamp_before: Optional[datetime.datetime] = None,
        timestamp_after: Optional[datetime.datetime] = None,
        from_state: Optional[str] = None,
        to_state: Optional[str] = None,
    ) -> EventLog:
        """
        Retrieve exactly one EventLog. Raises EventLog.DoesNotExist
        if no items were found, or EventLog.MultipleObjectsReturned if
        more than one item matched the query.

        id:               Only return EventLogs having an id in this list.
        job_id:           Only return Events associated with Job IDs in this list.
        batch_job_id:     Only return Events associated this BatchJob id.
        scheduler_id:     Only return Events associated with this HPC scheduler job ID.
        tags:             Only return Events for Jobs containing these tags (list of KEY:VALUE strings)
        data:             Only return Events containing this data (list of KEY:VALUE strings)
        timestamp_before: Only return Events before this time (UTC).
        timestamp_after:  Only return Events that occured after this time (UTC).
        from_state:       Only return Events transitioning from this Job state.
        to_state:         Only return Events transitioning to this Job state.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return EventLogQuery(manager=self).get(**kwargs)

    def filter(
        self,
        id: Union[typing.List[int], int, None] = None,
        job_id: Union[typing.List[int], int, None] = None,
        batch_job_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        tags: Union[typing.List[str], str, None] = None,
        data: Union[typing.List[str], str, None] = None,
        timestamp_before: Optional[datetime.datetime] = None,
        timestamp_after: Optional[datetime.datetime] = None,
        from_state: Optional[str] = None,
        to_state: Optional[str] = None,
    ) -> "EventLogQuery":
        """
        Returns a EventLog Query returning items matching the filter criteria.

        id:               Only return EventLogs having an id in this list.
        job_id:           Only return Events associated with Job IDs in this list.
        batch_job_id:     Only return Events associated this BatchJob id.
        scheduler_id:     Only return Events associated with this HPC scheduler job ID.
        tags:             Only return Events for Jobs containing these tags (list of KEY:VALUE strings)
        data:             Only return Events containing this data (list of KEY:VALUE strings)
        timestamp_before: Only return Events before this time (UTC).
        timestamp_after:  Only return Events that occured after this time (UTC).
        from_state:       Only return Events transitioning from this Job state.
        to_state:         Only return Events transitioning to this Job state.
        """
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"] and v is not None}
        return EventLogQuery(manager=self).filter(**kwargs)



import json
import sys
from typing import Any, Dict, List, Tuple, Type

from balsam._api.app import ApplicationDefinition, is_appdef
from balsam._api.models import App, Job
from balsam.schemas import SerializeError, serialize, serialize_exception


def is_mpi_rank_nonzero() -> bool:
    """Return True only if MPI is loaded and rank is >= 1"""
    MPI = None
    if "mpi4py" in sys.modules:
        MPI = getattr(sys.modules["mpi4py"], "MPI", None)

    if MPI is None:
        return False

    try:
        return MPI.COMM_WORLD.Get_rank() > 0  # type: ignore
    except Exception as exc:
        print(f"WARNING Get_rank() failed: {exc}")
        return False


def unpack_chunks(
    app_id: int, num_app_chunks: int, chunks: List[str]
) -> Tuple[Type[ApplicationDefinition], Job, Dict[str, Any]]:
    # sys.argv contains the serialized ApplicationDefinition and app_id (not the JSON app representation)
    app_encoded = "".join(chunks[:num_app_chunks])
    app = App(_api_data=True, id=app_id, site_id=0, name="AppName", serialized_class=app_encoded, source_code="")
    app_def = ApplicationDefinition.from_serialized(app)
    if not is_appdef(app_def):
        raise ValueError(f"{app_def} is not an ApplicationDefinition; type is {type(app_def)}")

    # sys.argv contains the full JSON representation of the Job:
    job_encoded = "".join(chunks[num_app_chunks:])
    job_dict: Dict[str, Any] = json.loads(job_encoded)
    job = Job._from_api(job_dict)
    params = job.get_parameters()

    return app_def, job, params


def log_exception(exc: Exception) -> None:
    if is_mpi_rank_nonzero():
        return

    try:
        serialized_exception = serialize_exception(exc)
    except SerializeError as ser_exc:
        print(f"Warning: failed to serialize the original exception that occured: {exc}")
        print(f"This was the exception-serialization error: {ser_exc}")
    else:
        print("BALSAM-EXCEPTION", serialized_exception, flush=True)


def log_result(ret_val: Any) -> None:
    if is_mpi_rank_nonzero():
        return
    print("BALSAM-RETURN-VALUE", serialize(ret_val), flush=True)


def main(app_id: int, num_app_chunks: int, chunks: List[str]) -> None:
    try:
        app_cls, job, params = unpack_chunks(app_id, num_app_chunks, chunks)
        app = app_cls(job)
        if not callable(app.run):
            raise AttributeError(f"ApplicationDefinition {app_cls} does not have a run() function")
        return_value = app.run(**params)
        log_result(return_value)
    except Exception as exc:
        log_exception(exc)
        raise


if __name__ == "__main__":
    # python -m balsam.site.launcher.python_runner APP_ID NUM_APP_CHUNKS [APP_B64_CHUNKS] [JOB_JSON_CHUNKS]
    app_id = int(sys.argv[1])
    num_app_chunks = int(sys.argv[2])
    main(app_id, num_app_chunks, sys.argv[3:])



from typing import List

from fastapi import APIRouter

from balsam.server import settings
from balsam.server.conf import LoginMethod

from . import authorization_code_login, device_code_login, password_login
from .token import user_from_token

LOGIN_ROUTERS = {
    LoginMethod.oauth_authcode: authorization_code_login.router,
    LoginMethod.oauth_device: device_code_login.router,
    LoginMethod.password: password_login.router,
}


def build_auth_router() -> APIRouter:
    """
    Enable login routes based on the configuration `settings.auth.login_methods`.
    Deployment config can thus enable/disable Oauth, Device flow, Password APIs
    """
    auth_router = APIRouter()
    for method in settings.auth.login_methods:
        auth_router.include_router(LOGIN_ROUTERS[method])

    @auth_router.get("/how")
    def get_auth_methods() -> List[str]:
        methods = [str(s) for s in settings.auth.login_methods]
        return methods

    return auth_router


__all__ = [
    "user_from_token",
    "build_auth_router",
]



import json
import logging
import os
import shlex
import subprocess
import sys
from datetime import datetime
from importlib.util import find_spec
from pathlib import Path
from typing import Any, Dict, List, Union

import click

from balsam.cmdline.utils import load_site_config, validate_partitions, validate_tags
from balsam.config import SiteConfig
from balsam.platform.app_run import AppRun
from balsam.platform.compute_node import ComputeNode
from balsam.schemas import BatchJobPartition
from balsam.site.launcher import NodeSpec
from balsam.util import SigHandler

MPI_MODE_PATH = find_spec("balsam.site.launcher.mpi_mode").origin  # type: ignore
SERIAL_MODE_PATH = find_spec("balsam.site.launcher.serial_mode").origin  # type: ignore
PART_INDEX = 0
logger = logging.getLogger("balsam.cmdline.launcher")


def get_run_basename(base: str) -> str:
    global PART_INDEX
    timestamp = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    fname = f"{base}_{timestamp}.{PART_INDEX}"
    PART_INDEX += 1
    return fname


def start_mpi_mode(
    site_config: SiteConfig, wall_time_min: int, nodes: List[ComputeNode], filter_tags: Dict[str, str]
) -> "subprocess.Popen[bytes]":
    basename = get_run_basename("mpi_mode")
    log_filename = site_config.log_path.joinpath(basename + ".log")
    stdout_filename = site_config.log_path.joinpath(basename + ".out")
    node_ids = [n.node_id for n in nodes]

    env = os.environ.copy()
    env["BALSAM_LAUNCHER_NODES"] = str(len(nodes))
    env["BALSAM_JOB_MODE"] = "mpi"

    args = ""
    args += f"{sys.executable} {MPI_MODE_PATH} "
    args += f"--wall-time-min {wall_time_min} "
    args += f"--log-filename {log_filename} "
    args += f"--node-ids {shlex.quote(json.dumps(node_ids))} "
    args += f"--filter-tags {shlex.quote(json.dumps(filter_tags))} "

    proc = subprocess.Popen(
        args=shlex.split(args),
        env=env,
        stdout=open(stdout_filename, "wb"),
        stderr=subprocess.STDOUT,
    )
    logger.info(f"Started MPI mode launcher: {str(proc.args)}")
    return proc


def start_serial_mode(
    site_config: SiteConfig, wall_time_min: int, nodes: List[ComputeNode], filter_tags: Dict[str, str]
) -> AppRun:
    master_host = nodes[0].hostname
    master_port = 19876
    basename = get_run_basename("serial_mode")
    log_filename = site_config.log_path.joinpath(basename + ".log")
    stdout_filename = site_config.log_path.joinpath(basename + ".out")
    node_spec = NodeSpec(
        node_ids=[n.node_id for n in nodes],
        hostnames=[n.hostname for n in nodes],
    )
    env = os.environ.copy()
    env["BALSAM_LAUNCHER_NODES"] = str(len(nodes))
    env["BALSAM_JOB_MODE"] = "serial"

    args = ""
    args += f"{sys.executable} {SERIAL_MODE_PATH} "
    args += f"--wall-time-min {wall_time_min} "
    args += f"--master-address {master_host}:{master_port} "
    args += f"--log-filename {log_filename} "
    args += f"--num-workers {len(nodes)} "
    args += f"--filter-tags {shlex.quote(json.dumps(filter_tags))} "

    app_run = site_config.settings.launcher.mpi_app_launcher
    app = app_run(
        cmdline=args,
        preamble=None,
        envs=env,
        cwd=Path.cwd(),
        outfile_path=stdout_filename,
        node_spec=node_spec,
        ranks_per_node=1,
        threads_per_rank=1,
        threads_per_core=1,
        launch_params=site_config.settings.launcher.serial_mode_startup_params,
        gpus_per_rank=len(nodes[0].gpu_ids),
    )
    app.start()
    logger.info(f"Started Serial mode launcher: {app._cmdline}")
    return app


@click.command()
@click.option("-j", "--job-mode", type=click.Choice(["serial", "mpi"]), required=True)
@click.option("-tag", "--tag", "filter_tags", multiple=True, callback=validate_tags)
@click.option("-p", "--part", "partitions", multiple=True, callback=validate_partitions)
@click.option("-t", "--wall-time-min", required=True, type=int)
def launcher(
    job_mode: str,
    filter_tags: Dict[str, str],
    partitions: List[BatchJobPartition],
    wall_time_min: int,
) -> None:
    site_config = load_site_config()
    node_cls = site_config.settings.launcher.compute_node
    nodes = node_cls.get_job_nodelist()

    if not partitions:
        partitions = [BatchJobPartition(num_nodes=len(nodes), job_mode=job_mode, filter_tags=filter_tags)]

    partition_dicts: List[Dict[str, Any]] = [part.dict() for part in partitions]
    assert sum(p["num_nodes"] for p in partition_dicts) <= len(nodes)
    idx = 0
    for part in partition_dicts:
        num_nodes = part.pop("num_nodes")
        start, end = idx, idx + num_nodes
        part["nodes"] = nodes[start:end]
        idx += num_nodes

    sig_handler = SigHandler()

    launcher_procs = []
    for part in partition_dicts:
        job_mode = part.pop("job_mode")
        proc: "Union[subprocess.Popen[bytes], AppRun]"
        if job_mode == "mpi":
            proc = start_mpi_mode(
                site_config=site_config,
                wall_time_min=wall_time_min,
                **part,
            )
        else:
            proc = start_serial_mode(
                site_config=site_config,
                wall_time_min=wall_time_min,
                **part,
            )
        launcher_procs.append(proc)

    while launcher_procs and not sig_handler.is_set():
        done_procs = []
        for proc in launcher_procs:
            try:
                proc.wait(timeout=1)
            except (subprocess.TimeoutExpired):
                pass
            else:
                done_procs.append(proc)
        launcher_procs = [proc for proc in launcher_procs if proc not in done_procs]

    for proc in launcher_procs:
        proc.terminate()
    for proc in launcher_procs:
        try:
            proc.wait(timeout=10)
        except subprocess.TimeoutExpired:
            proc.kill()



import json
from typing import Any, Dict

from fastapi import APIRouter, Depends, status
from sqlalchemy import orm

from balsam import schemas
from balsam.server import models, settings
from balsam.server.models import crud, get_session
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import AppQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedAppsOut)
def list(
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[models.App] = Depends(Paginator),
    q: AppQuery = Depends(AppQuery),
) -> Dict[str, Any]:
    """List Apps registered by the user's Balsam Sites."""
    count, apps = crud.apps.fetch(
        db,
        owner=user,
        paginator=paginator,
        filterset=q,
    )
    return {"count": count, "results": apps}


@router.get("/{app_id}", response_model=schemas.AppOut)
def read(
    app_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> "orm.Query[models.App]":
    _, app = crud.apps.fetch(db, owner=user, app_id=app_id)
    """Get the specified App."""
    return app


@router.post("/", response_model=schemas.AppOut, status_code=status.HTTP_201_CREATED)
def create(
    app: schemas.AppCreate, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> schemas.AppOut:
    """
    Register a new Balsam App.
    """
    new_app = crud.apps.create(db, owner=user, app=app)
    result = schemas.AppOut.from_orm(new_app)
    db.commit()
    pubsub.publish(user.id, "create", "app", result)
    return result


@router.put("/{app_id}", response_model=schemas.AppOut)
def update(
    app_id: int, app: schemas.AppUpdate, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> models.App:
    """Update an App."""
    data = json.loads(app.json(exclude_unset=True))
    updated_app = crud.apps.update(db, owner=user, app_id=app_id, update_data=data)
    data["id"] = app_id
    db.commit()
    pubsub.publish(user.id, "update", "app", data)
    return updated_app


@router.delete("/{app_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(app_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> None:
    """Delete an App and all associated Jobs are deleted."""
    crud.apps.delete(db, owner=user, app_id=app_id)
    db.commit()
    pubsub.publish(user.id, "delete", "app", {"id": app_id})



import itertools
import logging
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, DefaultDict, List

from balsam._api.app import ApplicationDefinition
from balsam.schemas import DeserializeError, JobState

from .service_base import BalsamService

if TYPE_CHECKING:
    from balsam._api.models import Job  # noqa: F401
    from balsam.client import RESTClient

logger = logging.getLogger(__name__)


class FileCleanerService(BalsamService):
    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        apps_path: Path,
        data_path: Path,
        cleanup_batch_size: int = 180,  # cleanup after 180 jobs
        service_period: int = 30,  # cleanup every 30 seconds
    ) -> None:
        super().__init__(client=client, service_period=service_period)
        self.site_id = site_id
        self.data_path = data_path
        self.cleanup_batch_size = cleanup_batch_size
        ApplicationDefinition._set_client(client)
        try:
            ApplicationDefinition.load_by_site(self.site_id)  # Warms the cache
        except DeserializeError as exc:
            logger.warning(
                f"At least one App registered at this Site failed to deserialize: {exc}. "
                "Jobs running with this App will not have their files cleaned up. Please fix the App classes and/or the "
                "Balsam Site environment, and double check that the Apps can load OK."
            )

    def remove_files(self, jobs: List["Job"], cleanup_file_patterns: List[str]) -> None:
        globs = itertools.chain(
            *(job.resolve_workdir(self.data_path).glob(pattern) for job in jobs for pattern in cleanup_file_patterns)
        )
        for match in globs:
            pstr = match.as_posix()
            if pstr.startswith(self.data_path.as_posix()) and ".." not in pstr:
                try:
                    match.unlink()
                    logger.info(f"Removed {pstr}")
                except (OSError, FileNotFoundError) as exc:
                    logger.warning(f"Cannot remove {pstr}: {exc}")

    def run_cycle(self) -> None:
        jobs_by_appid: DefaultDict[int, List["Job"]] = defaultdict(list)
        qs = self.client.Job.objects.filter(
            site_id=self.site_id, state=JobState.job_finished, pending_file_cleanup=True
        )
        for job in qs[: self.cleanup_batch_size]:
            jobs_by_appid[job.app_id].append(job)

        for appid, jobs in jobs_by_appid.items():
            try:
                app_cls = ApplicationDefinition.load_by_id(appid)
            except DeserializeError as exc:
                logger.error(f"Failed to deserialize App {appid}; will not clean up after these jobs: {exc}")
            else:
                cleanup_file_patterns = app_cls.cleanup_files
                self.remove_files(jobs, cleanup_file_patterns)
            for job in jobs:
                job.pending_file_cleanup = False
            self.client.Job.objects.bulk_update(jobs)

    def cleanup(self) -> None:
        logger.info("Exiting FileCleaner service")



from datetime import datetime
from typing import Iterable, List, Optional, Sequence, Tuple, cast

from sqlalchemy import orm
from sqlalchemy.orm import Query, Session

from balsam import schemas
from balsam.server import ValidationError, models
from balsam.server.routers.filters import TransferItemQuery
from balsam.server.utils import Paginator


def owned_transfer_query(db: Session, owner: schemas.UserOut) -> "Query[models.TransferItem]":
    return cast(
        "Query[models.TransferItem]",
        db.query(models.TransferItem)
        .join(models.Job)  # type: ignore
        .join(models.App)
        .join(models.Site)
        .filter(models.Site.owner_id == owner.id),
    )


def fetch(
    db: Session,
    owner: schemas.UserOut,
    paginator: Optional[Paginator[models.TransferItem]] = None,
    transfer_id: Optional[int] = None,
    filterset: Optional[TransferItemQuery] = None,
) -> Tuple[int, Iterable[models.TransferItem]]:
    qs = owned_transfer_query(db, owner)
    if transfer_id is not None:
        qs = qs.filter(models.TransferItem.id == transfer_id)
        transfer_item = qs.one()
        return 1, [transfer_item]
    if filterset is not None:
        qs = filterset.apply_filters(qs)
    count = qs.group_by(models.TransferItem.id).count()
    assert paginator is not None
    transfers = paginator.paginate(qs)
    return count, transfers


def _set_transfer_state(job: models.Job) -> bool:
    direction = "in" if job.state == "READY" else "out"
    is_done = all(item.state == "done" for item in job.transfer_items if item.direction == direction)
    if is_done:
        job.state = "STAGED_IN" if direction == "in" else "JOB_FINISHED"
    return is_done


def _update_jobs(db: Session, job_ids: Sequence[int]) -> Tuple[List[models.Job], List[models.LogEvent]]:
    jobs = (
        db.query(models.Job)
        .filter(models.Job.id.in_(job_ids), models.Job.state.in_(["READY", "POSTPROCESSED"]))
        .options(  # type: ignore
            orm.selectinload(models.Job.transfer_items).load_only(
                models.TransferItem.direction,
                models.TransferItem.state,
            ),
            orm.selectinload(models.Job.parents).load_only(models.Job.id),
        )
        .all()
    )
    updates, events = [], []
    now = datetime.utcnow()
    for job in jobs:
        job.parent_ids = [parent.id for parent in job.parents]
        old_state = job.state
        if _set_transfer_state(job):
            event = models.LogEvent(
                job=job,
                timestamp=now,
                from_state=old_state,
                to_state=job.state,
            )
            updates.append(job)
            events.append(event)
    db.flush()
    return updates, events


def update(
    db: Session, owner: schemas.UserOut, transfer_id: int, data: schemas.TransferItemUpdate
) -> Tuple[models.TransferItem, Optional[models.Job], Optional[models.LogEvent]]:
    item = owned_transfer_query(db, owner).filter(models.TransferItem.id == transfer_id).one()
    for k, v in data.dict(exclude_unset=True).items():
        setattr(item, k, v)
    db.flush()
    updated_jobs, events = _update_jobs(db, [item.job_id])
    updated_job = updated_jobs[0] if updated_jobs else None
    event = events[0] if events else None
    return item, updated_job, event


def bulk_update(
    db: Session, owner: schemas.UserOut, update_list: List[schemas.TransferItemBulkUpdate]
) -> Tuple[List[models.TransferItem], List[models.Job], List[models.LogEvent]]:
    updates = {item.id: item.dict(exclude_unset=True) for item in update_list}
    db_items = owned_transfer_query(db, owner).filter(models.TransferItem.id.in_(updates.keys())).all()
    if len(db_items) < len(updates):
        raise ValidationError("Some TransferItems not found")
    for transfer_item in db_items:
        for k, v in updates[transfer_item.id].items():
            setattr(transfer_item, k, v)
    db.flush()
    updated_jobs, events = _update_jobs(db, [item.job_id for item in db_items])
    return db_items, updated_jobs, events



from datetime import datetime
from typing import Any, Dict, Optional, Tuple, Union

from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Query, Session

from balsam import schemas
from balsam.server import ValidationError, models
from balsam.server.routers.filters import SiteQuery
from balsam.server.utils import Paginator


def fetch(
    db: Session,
    owner: schemas.UserOut,
    paginator: Optional[Paginator[models.Site]] = None,
    site_id: Optional[int] = None,
    filterset: Optional[SiteQuery] = None,
) -> "Tuple[int, Union[models.Site, Query[models.Site]]]":
    qs = db.query(models.Site).filter(models.Site.owner_id == owner.id)
    if filterset is not None:
        qs = filterset.apply_filters(qs)

    if site_id is not None:
        site = qs.filter(models.Site.id == site_id).one()
        return (1, site)
    else:
        assert paginator is not None
        count = qs.group_by(models.Site.id).count()
        return count, paginator.paginate(qs)


def create(db: Session, owner: schemas.UserOut, site: schemas.SiteCreate) -> models.Site:
    site_id = db.query(models.Site.id).filter_by(name=site.name, owner_id=owner.id).scalar()  # type: ignore
    if site_id is not None:
        raise ValidationError("A site with this name already exists")
    new_site = models.Site(
        **jsonable_encoder(site),
        creation_date=datetime.utcnow(),
        last_refresh=datetime.utcnow(),
        owner_id=owner.id,
    )
    db.add(new_site)
    db.flush()
    return new_site


def update(db: Session, owner: schemas.UserOut, site_id: int, update_data: Dict[str, Any]) -> models.Site:
    qs = db.query(models.Site).filter(models.Site.owner_id == owner.id, models.Site.id == site_id)
    site_in_db = qs.one()
    qs.update(jsonable_encoder(update_data))
    db.flush()
    return site_in_db


def delete(db: Session, owner: schemas.UserOut, site_id: int) -> None:
    site_in_db = db.query(models.Site).filter(models.Site.owner_id == owner.id, models.Site.id == site_id).one()
    db.delete(site_in_db)  # type: ignore
    db.flush()



from datetime import datetime
from typing import Optional

import dateutil.parser
import dateutil.tz

et = dateutil.tz.gettz("America/New York")
ct = dateutil.tz.gettz("America/Chicago")
mt = dateutil.tz.gettz("America/Denver")
pt = dateutil.tz.gettz("America/Los Angeles")

TZINFOS = {
    "ET": et,
    "EST": et,
    "EDT": et,
    "CT": ct,
    "CST": ct,
    "CDT": ct,
    "MT": mt,
    "MST": mt,
    "MDT": mt,
    "PT": pt,
    "PST": pt,
    "PDT": pt,
}


def parse_to_utc(time_str: str, local_zone: Optional[str] = None) -> datetime:
    """
    Converts a non-UTC, timezone-aware string to a UTC datetime object
    """
    local_tz = {"L": TZINFOS[local_zone]} if local_zone else {}
    dt = dateutil.parser.parse(time_str, tzinfos={**TZINFOS, **local_tz})
    dt_utc = dt.astimezone(dateutil.tz.UTC)
    return dt_utc



"""APIClient-driven tests"""
import random
import time
from datetime import datetime, timedelta
from uuid import uuid4

import pytest
from dateutil.parser import isoparse
from fastapi import status

from balsam.server import models

from .util import create_app, create_site

FAST_EXPIRATION_PERIOD = 0.6


@pytest.fixture(scope="function")
def site(auth_client):
    """Called before each test"""
    return create_site(
        auth_client,
        name="site1",
        transfer_locations={"MyCluster": f"globus://{uuid4()}"},
    )


@pytest.fixture(scope="function")
def app(auth_client, site):
    return create_app(auth_client, site_id=site["id"], name="Hello")


@pytest.fixture(scope="function")
def job_dict(app):
    def _job_dict(
        app_id=None,
        workdir="test/1",
        tags={},
        transfers={"hello-input": {"location_alias": "MyCluster", "path": "/path/to/input.dat"}},
        parameters={"name": "world", "N": 4},
        data={},
        return_code=None,
        parent_ids=[],
        num_nodes=2,
        ranks_per_node=4,
        threads_per_rank=1,
        threads_per_core=1,
        launch_params={},
        gpus_per_rank=0,
        node_packing_count=1,
        wall_time_min=0,
    ):
        if app_id is None:
            app_id = app["id"]
        return dict(
            app_id=app_id,
            workdir=workdir,
            tags=tags,
            transfers=transfers,
            parameters=parameters,
            data=data,
            return_code=return_code,
            parent_ids=parent_ids,
            num_nodes=num_nodes,
            ranks_per_node=ranks_per_node,
            threads_per_rank=threads_per_rank,
            threads_per_core=threads_per_core,
            launch_params=launch_params,
            gpus_per_rank=gpus_per_rank,
            node_packing_count=node_packing_count,
            wall_time_min=wall_time_min,
        )

    return _job_dict


@pytest.fixture(scope="function")
def create_session(job_dict, site, db_session):
    def _create_session():
        bjob = models.BatchJob(
            site_id=site["id"],
            scheduler_id=123,
            project="foo",
            queue="default",
            state="running",
            num_nodes=32,
            wall_time_min=60,
            job_mode="mpi",
            filter_tags={},
        )
        db_session.add(bjob)
        session = models.Session(site_id=bjob.site_id, batch_job=bjob)
        db_session.commit()
        return session

    return _create_session


@pytest.fixture(scope="function")
def fast_session_expiry():
    old_expiry = models.crud.sessions.SESSION_EXPIRE_PERIOD
    old_sweep = models.crud.sessions.SESSION_SWEEP_PERIOD
    try:
        models.crud.sessions.SESSION_EXPIRE_PERIOD = timedelta(seconds=FAST_EXPIRATION_PERIOD)
        models.crud.sessions.SESSION_SWEEP_PERIOD = timedelta(seconds=0.1)
        yield
    finally:
        models.crud.sessions.SESSION_EXPIRE_PERIOD = old_expiry
        models.crud.sessions.SESSION_SWEEP_PERIOD = old_sweep


def assertHistory(client, job, *states, **expected_messages):
    """
    Assert that `job` went through the sequence of `states` in order.
    For each state:str pair in `expected_messages`, verify the str is contained
    in the transition log message.
    """
    response = client.get("/events", job_id=job["id"])
    fail_msg = "\n" + "\n".join(
        f'{i}) {e["from_state"]} ->  {e["to_state"]} ({e["data"]})' for i, e in enumerate(response["results"])
    )
    assert response["count"] == len(states) - 1, fail_msg
    eventlogs = list(reversed(response["results"]))

    for i, (from_state, to_state) in enumerate(zip(states[:-1], states[1:])):
        expected_dict = {"from_state": from_state, "to_state": to_state}
        event = eventlogs[i]
        actual = {key: event[key] for key in ("from_state", "to_state")}
        assert expected_dict == actual
        if to_state in expected_messages:
            assert expected_messages.pop(to_state) in event["data"]["message"]


@pytest.fixture(scope="function")
def linear_dag(auth_client, job_dict):
    A = auth_client.bulk_post("/jobs/", [job_dict(tags={"step": "A", "dag": "dag1"})])[0]
    B = auth_client.bulk_post("/jobs/", [job_dict(tags={"step": "B", "dag": "dag1"}, parent_ids=[A["id"]])])[0]
    C = auth_client.bulk_post("/jobs/", [job_dict(tags={"step": "C", "dag": "dag1"}, parent_ids=[B["id"]])])[0]
    return A, B, C


def test_add_jobs(auth_client, job_dict):
    jobs = [
        job_dict(
            parameters={"name": "foo", "N": i},
            workdir=f"test/{i}",
        )
        for i in range(3)
    ]
    jobs = auth_client.bulk_post("/jobs/", jobs)
    for job in jobs:
        assert job["state"] == "READY"
        assertHistory(auth_client, job, "CREATED", "READY")


def test_child_with_two_parents_state_update(auth_client, job_dict):
    resp = auth_client.bulk_post("/jobs/", [job_dict()])
    parent1 = resp[0]
    resp = auth_client.bulk_post("/jobs/", [job_dict()])
    parent2 = resp[0]
    resp = auth_client.bulk_post("/jobs/", [job_dict(parent_ids=[parent1["id"], parent2["id"]])])
    child = resp[0]
    assert child["state"] == "AWAITING_PARENTS"

    auth_client.bulk_patch("/jobs/", [{"id": parent1["id"], "state": "JOB_FINISHED"}])
    child = auth_client.get(f"/jobs/{child['id']}")
    assert child["state"] == "AWAITING_PARENTS"

    auth_client.bulk_patch("/jobs/", [{"id": parent2["id"], "state": "JOB_FINISHED"}])
    child = auth_client.get(f"/jobs/{child['id']}")
    assert child["state"] == "READY"


def test_parent_with_two_children_state_update(auth_client, job_dict):
    resp = auth_client.bulk_post("/jobs/", [job_dict()])
    parent = resp[0]
    resp = auth_client.bulk_post("/jobs/", [job_dict(parent_ids=[parent["id"]])])
    child1 = resp[0]
    resp = auth_client.bulk_post("/jobs/", [job_dict(parent_ids=[parent["id"]])])
    child2 = resp[0]

    assert parent["state"] == "READY"
    assert child1["state"] == "AWAITING_PARENTS" == child2["state"]

    # POSTPROCESSED Job cascades to JOB_FINISHED:
    auth_client.bulk_put("/jobs/", {"state": "POSTPROCESSED"}, id=parent["id"])
    child1 = auth_client.get(f"/jobs/{child1['id']}")
    child2 = auth_client.get(f"/jobs/{child2['id']}")
    assert child1["state"] == "READY"
    assert child2["state"] == "READY"


def test_add_job_without_transfers_is_STAGED_IN(auth_client, job_dict):
    """Ready and bound to backend"""
    job = auth_client.bulk_post("/jobs/", [job_dict(transfers={})])[0]
    assert job["state"] == "STAGED_IN"


def test_bulk_put(auth_client, job_dict):
    jobs = auth_client.bulk_post("/jobs/", [job_dict(transfers={}) for _ in range(10)])
    for job in jobs:
        assert job["state"] == "STAGED_IN"
    ids = [j["id"] for j in jobs]
    jobs = auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"}, id=ids)
    for job in jobs:
        assert job["state"] == "PREPROCESSED"


def test_bulk_patch(auth_client, job_dict):
    jobs = auth_client.bulk_post("/jobs/", [job_dict(transfers={}) for _ in range(10)])
    for job in jobs:
        assert job["state"] == "STAGED_IN"
    ids = [j["id"] for j in jobs]
    jobs = auth_client.bulk_patch("/jobs/", [{"id": id, "state": "PREPROCESSED"} for id in ids])
    for job in jobs:
        assert job["state"] == "PREPROCESSED"


def test_acquire_for_launch(auth_client, job_dict, create_session):
    """Jobs become associated with BatchJob"""
    jobs = auth_client.bulk_post("/jobs/", [job_dict(transfers={}) for _ in range(10)])
    for job in jobs:
        assert job["state"] == "STAGED_IN"
        assert job["batch_job_id"] is None

    # Mark jobs PREPROCESSED and ready to run
    ids = [j["id"] for j in jobs]
    app_ids = {jobs[0]["app_id"]}
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"}, id=ids)

    session1, session2 = create_session(), create_session()

    # Launcher1 acquires 2 runnable jobs:
    acquired1 = auth_client.post(
        f"/sessions/{session1.id}",
        max_wall_time_min=120,
        filter_tags={},
        max_num_jobs=2,
        max_nodes_per_job=32,
        app_ids=app_ids,
        check=status.HTTP_200_OK,
    )

    # Launcher2 asks for up to 1000 and gets all the rest:
    acquired2 = auth_client.post(
        f"/sessions/{session2.id}",
        max_wall_time_min=120,
        filter_tags={},
        max_num_jobs=1000,
        max_nodes_per_job=32,
        app_ids=app_ids,
        check=status.HTTP_200_OK,
    )

    # These jobs are now associated with the corresponding BatchJob
    assert len(acquired1) == 2
    assert len(acquired2) == 8
    for job in acquired1:
        assert job["batch_job_id"] == session1.batch_job_id
    for job in acquired2:
        assert job["batch_job_id"] == session2.batch_job_id


def test_update_to_running_does_not_release_lock(auth_client, job_dict, create_session, db_session):
    jobs = auth_client.bulk_post("/jobs/", [job_dict(transfers={}) for _ in range(10)])

    # Mark jobs PREPROCESSED
    ids = [j["id"] for j in jobs]
    app_ids = {jobs[0]["app_id"]}
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"}, id=ids)

    session = create_session()
    # Launcher1 acquires all of them
    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=120,
        filter_tags={},
        max_num_jobs=100,
        max_nodes_per_job=32,
        app_ids=app_ids,
        check=status.HTTP_200_OK,
    )

    # Behind the scenes, the acquired jobs are locked:
    for job in db_session.query(models.Job).all():
        assert job.session_id == session.id

    # The jobs start running in a staggered fashion; a bulk status update is made
    run_start_times = [datetime.utcnow() + timedelta(seconds=random.randint(0, 20)) for _ in acquired]
    updates = [
        {
            "id": j["id"],
            "state": "RUNNING",
            "state_timestamp": ts,
            "state_message": "Running on Theta nid00139",
        }
        for j, ts in zip(acquired, run_start_times)
    ]
    jobs = auth_client.bulk_patch("/jobs/", updates)

    # The jobs are associated to batchjob and have the expected history:
    for job in jobs:
        assert job["batch_job_id"] == session.batch_job_id
        assertHistory(auth_client, job, "CREATED", "STAGED_IN", "PREPROCESSED", "RUNNING")

    # Behind the scenes, the acquired jobs have changed state and are still locked:
    for job in db_session.query(models.Job).all():
        assert job.session_id == session.id

    # The EventLogs were correctly recorded:
    time_stamps_in_db = (
        db_session.query(models.LogEvent.timestamp).filter(models.LogEvent.to_state == "RUNNING").all()
    )
    time_stamps_in_db = set(tup[0] for tup in time_stamps_in_db)
    assert time_stamps_in_db == set(run_start_times)


def test_acquire_for_launch_with_node_constraints(auth_client, job_dict, create_session):
    jobs = [
        *[job_dict(num_nodes=1, ranks_per_node=1, wall_time_min=40) for _ in range(2)],
        *[job_dict(num_nodes=1, ranks_per_node=1, wall_time_min=30) for _ in range(2)],
        *[job_dict(num_nodes=1, ranks_per_node=1, node_packing_count=4, wall_time_min=30) for _ in range(4)],
        *[job_dict(num_nodes=1, ranks_per_node=1, wall_time_min=50) for _ in range(2)],
        *[job_dict(num_nodes=1, ranks_per_node=4, wall_time_min=30) for _ in range(2)],
        *[job_dict(num_nodes=3, wall_time_min=0) for _ in range(2)],
        *[job_dict(num_nodes=8, wall_time_min=0) for _ in range(2)],
        *[job_dict(num_nodes=16, wall_time_min=0) for _ in range(2)],
    ]
    jobs = auth_client.bulk_post("/jobs/", jobs)
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"}, id=[j["id"] for j in jobs])
    session = create_session()

    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=32,
        max_nodes_per_job=1,
        serial_only=True,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 8
    assert all(job["num_nodes"] == 1 for job in acquired)
    assert all(job["ranks_per_node"] == 1 for job in acquired)
    assert acquired == sorted(acquired, key=lambda job: (job["node_packing_count"], -1 * job["wall_time_min"]))

    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=100,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 6
    assert all(1 <= job["num_nodes"] <= 8 for job in acquired)


def test_acquire_by_tags(auth_client, job_dict, create_session):
    jobs = [
        *[job_dict(tags={"system": "H2O", "calc": "energy"}) for _ in range(3)],
        *[job_dict(tags={"system": "H2O", "calc": "vib"}) for _ in range(3)],
        *[job_dict(tags={"system": "D2O", "calc": "energy"}) for _ in range(3)],
    ]
    jobs = auth_client.bulk_post("/jobs/", jobs)
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"}, id=[j["id"] for j in jobs])
    session = create_session()

    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={"system": "T2O"},
        max_nodes_per_job=8,
        max_num_jobs=8,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 0

    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={"system": "H2O", "calc": "vib"},
        max_nodes_per_job=8,
        max_num_jobs=8,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 3
    assert all(job["tags"]["system"] == "H2O" for job in acquired)
    assert all(job["tags"]["calc"] == "vib" for job in acquired)

    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={"calc": "energy"},
        max_nodes_per_job=8,
        max_num_jobs=8,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 6
    assert all(job["tags"]["calc"] == "energy" for job in acquired)


def test_bulk_update_based_on_tags_filter_via_put(auth_client, job_dict):
    specs = [
        job_dict(tags={"mass": 1.0}, num_nodes=1),
        job_dict(tags={"mass": 1.0}, num_nodes=1),
        job_dict(tags={"mass": 2.0}, num_nodes=1),
        job_dict(tags={"mass": 2.0}, num_nodes=1),
    ]
    auth_client.bulk_post("/jobs/", specs)
    auth_client.bulk_put(
        "/jobs/",
        {"num_nodes": 128},
        tags=["mass:2.0"],
    )
    jobs = auth_client.get("/jobs/")["results"]
    assert len(jobs) == 4
    for j in jobs:
        if j["tags"]["mass"] == "1.0":
            assert j["num_nodes"] == 1
        else:
            assert j["tags"]["mass"] == "2.0"
            assert j["num_nodes"] == 128


def test_can_filter_on_id(auth_client, job_dict):
    specs = [
        job_dict(workdir="A"),
        job_dict(workdir="B"),
        job_dict(workdir="C"),
    ]
    A, B, C = auth_client.bulk_post("/jobs/", specs)
    res = auth_client.get("/jobs/", id=[B["id"], C["id"]], ordering="workdir")
    assert res["count"] == 2
    workdirs = [job["workdir"] for job in res["results"]]
    assert workdirs == ["B", "C"]


def test_can_filter_on_parents(auth_client, job_dict):
    specs = [job_dict(workdir="A"), job_dict(workdir="B")]
    parentA, parentB = auth_client.bulk_post("/jobs/", specs)

    child_specs = [
        job_dict(workdir="A1", parent_ids=[parentA["id"]]),
        job_dict(workdir="A2", parent_ids=[parentA["id"]]),
        job_dict(workdir="B1", parent_ids=[parentB["id"]]),
        job_dict(workdir="B2", parent_ids=[parentB["id"]]),
        job_dict(workdir="B3", parent_ids=[parentB["id"]]),
        job_dict(workdir="C1", parent_ids=[parentA["id"], parentB["id"]]),
    ]
    auth_client.bulk_post("/jobs/", child_specs)

    children_of_B = auth_client.get(
        "/jobs/",
        parent_id=[parentB["id"]],
        ordering="workdir",
    )
    assert children_of_B["count"] == 4
    workdirs = [job["workdir"] for job in children_of_B["results"]]
    assert workdirs == ["B1", "B2", "B3", "C1"]

    children = auth_client.get(
        "/jobs/",
        parent_id=[parentA["id"], parentB["id"]],
        ordering="workdir",
    )
    assert children["count"] == 6
    workdirs = [job["workdir"] for job in children["results"]]
    assert workdirs == ["A1", "A2", "B1", "B2", "B3", "C1"]


def test_can_filter_on_last_update(auth_client, job_dict):
    specs = [job_dict(workdir="A"), job_dict(workdir="B")]
    A, B = auth_client.bulk_post("/jobs/", specs)
    creation_time = datetime.utcnow()  # IMPORTANT! all times in UTC

    time.sleep(0.1)
    auth_client.bulk_patch("/jobs/", [{"id": B["id"], "state": "PREPROCESSED"}])

    # Before the creation_timestamp: only A
    jobs = auth_client.get("/jobs/", last_update_before=creation_time)
    assert jobs["count"] == 1
    assert jobs["results"][0]["workdir"] == "A"

    # After the creation_timestamp: only B
    jobs = auth_client.get("/jobs/", last_update_after=creation_time)
    assert jobs["count"] == 1
    assert jobs["results"][0]["workdir"] == "B"


def test_update_to_run_done_releases_lock_but_not_batch_job(auth_client, job_dict, create_session, db_session):
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(5)])
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})

    # Before acqusition: no locks, no batchjobs assigned
    for job in db_session.query(models.Job):
        assert job.session_id is None
        assert job.batch_job_id is None

    # Acquisition
    session = create_session()
    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_nodes_per_job=8,
        max_num_jobs=8,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 5

    # After acqusition: locks & batchjob assigned
    for job in db_session.query(models.Job):
        assert job.session_id == session.id
        assert job.batch_job_id == session.batch_job_id

    # Update all to RUNNING
    auth_client.bulk_put("/jobs/", {"state": "RUNNING"})

    # After RUNNING: locks & batchjob assigned
    db_session.expire_all()
    for job in db_session.query(models.Job):
        assert job.state == "RUNNING"
        assert job.session_id == session.id
        assert job.batch_job_id == session.batch_job_id

    # Update to RUN_DONE
    auth_client.bulk_put("/jobs/", {"state": "RUN_DONE"})

    # After RUN_DONE: locks freed; batchjob remains
    db_session.expire_all()
    for job in db_session.query(models.Job):
        assert job.state == "RUN_DONE"
        assert job.session_id is None
        assert job.batch_job_id == session.batch_job_id


def test_tick_heartbeat_extends_expiration(auth_client, job_dict, create_session, db_session):
    session = create_session()
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(5)])
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})

    before_acquire = datetime.utcnow()
    for job in db_session.query(models.Job):
        assert job.session_id is None
    db_session.expire_all()

    # Acquire session lock on all jobs
    auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_nodes_per_job=8,
        max_num_jobs=8,
        check=status.HTTP_200_OK,
    )

    db_session.refresh(session)
    after_acquire = session.heartbeat
    assert session.jobs.count() == 5
    assert before_acquire < after_acquire

    # Tick lock some time later
    time.sleep(0.15)
    auth_client.put(f"/sessions/{session.id}")

    db_session.refresh(session)
    after_tick = session.heartbeat
    assert (after_tick - after_acquire) > timedelta(seconds=0.1)


def test_clear_expired_sess(auth_client, job_dict, create_session, db_session, fast_session_expiry):
    session1, session2 = create_session(), create_session()
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(10)])
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})

    # Session1 acquires 5 jobs
    auth_client.post(
        f"/sessions/{session1.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=5,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )

    # Session2 acquires the other 5
    auth_client.post(
        f"/sessions/{session2.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=5,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )

    # All are locked
    assert db_session.query(models.Job).filter_by(session_id=session1.id).count() == 5
    assert db_session.query(models.Job).filter_by(session_id=session2.id).count() == 5

    time.sleep(FAST_EXPIRATION_PERIOD / 2)

    # Session2 ticks
    auth_client.put(f"/sessions/{session2.id}")
    time.sleep(FAST_EXPIRATION_PERIOD / 2)

    # Session2 ticks again
    auth_client.put(f"/sessions/{session2.id}")
    time.sleep(FAST_EXPIRATION_PERIOD / 2)

    # By now, session1 is cleared because it expired
    sess1_id = session1.id
    db_session.expire_all()
    assert db_session.query(models.Job).filter_by(session_id=sess1_id).count() == 0
    assert db_session.query(models.Job).filter_by(session_id=session2.id).count() == 5
    assert db_session.query(models.Session).count() == 1

    # If session1 tries to tick at this point, it will get a 404: lock is gone
    auth_client.put(f"/sessions/{sess1_id}", check=status.HTTP_404_NOT_FOUND)


def test_view_session_list(auth_client, job_dict, create_session):
    session = create_session()
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(5)])
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})

    before_acquire = datetime.utcnow()

    # Session acquires 5 jobs
    auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=5,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )
    after_acquire = datetime.utcnow()
    sessions = auth_client.get("/sessions/")["results"]
    assert len(sessions) == 1
    sess = sessions[0]
    assert sess["batch_job_id"] == session.batch_job_id
    assert isoparse(sess["heartbeat"]) > before_acquire
    assert isoparse(sess["heartbeat"]) < after_acquire


def test_delete_session_frees_lock_on_all_jobs(auth_client, job_dict, create_session, db_session):
    session1, session2 = create_session(), create_session()
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(10)])
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})

    # Session1 acquires 5 jobs
    auth_client.post(
        f"/sessions/{session1.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=5,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )
    # Session2 acquires 5 jobs
    auth_client.post(
        f"/sessions/{session2.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=5,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )
    assert db_session.query(models.Job).filter_by(session_id=session1.id).count() == 5
    assert db_session.query(models.Job).filter_by(session_id=session2.id).count() == 5

    # Session1 is deleted, leaving its jobs unlocked
    auth_client.delete(f"/sessions/{session1.id}")
    db_session.expire_all()
    unlocked = db_session.query(models.Job).filter(models.Job.session_id.is_(None))
    assert unlocked.count() == 5
    assert db_session.query(models.Job).filter_by(session_id=session2.id).count() == 5


def test_update_transfer_item(auth_client, job_dict, db_session):
    """Can update state, status_message, task_id"""
    job = auth_client.bulk_post(
        "/jobs/",
        [
            job_dict(
                transfers={
                    "hello-input": {
                        "location_alias": "MyCluster",
                        "path": "/path/to/input.dat",
                    }
                },
            )
        ],
    )[0]

    resp = auth_client.get("/transfers")
    assert resp["count"] == 1

    transfer_item = resp["results"][0]
    assert transfer_item["state"] == "pending"
    assert transfer_item["job_id"] == job["id"]
    assert transfer_item["direction"] == "in"
    assert transfer_item["local_path"] == "hello.yml"
    assert str(transfer_item["remote_path"]) == "/path/to/input.dat"
    assert transfer_item["task_id"] == ""
    assert transfer_item["transfer_info"] == {}

    # Update state and task id
    tid = transfer_item["id"]
    new_globus_task_id = uuid4().hex
    auth_client.put(f"/transfers/{tid}", state="active", task_id=new_globus_task_id)

    # TransferItem is indeed updated
    transfer = db_session.query(models.TransferItem).one()
    assert transfer.state == "active"
    assert transfer.task_id == new_globus_task_id


def test_finished_transfer_updates_job_state(auth_client, job_dict):
    """Can update state, status_message, task_id"""
    job = auth_client.bulk_post(
        "/jobs/",
        [
            job_dict(
                transfers={
                    "hello-input": {
                        "location_alias": "MyCluster",
                        "path": "/path/to/input.dat",
                    }
                },
            )
        ],
    )[0]

    assert job["state"] == "READY"
    transfer_item = auth_client.get("/transfers")["results"][0]
    tid = transfer_item["id"]
    auth_client.put(f"/transfers/{tid}", state="done", task_id=uuid4().hex)

    # After finishing stage-in, job was auto-marked STAGED_IN:
    job = auth_client.get(f"/jobs/{job['id']}")
    assert job["state"] == "STAGED_IN"


# Viewing State History
def test_aggregated_state_history(auth_client, job_dict):
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(10)])
    auth_client.bulk_put("/jobs/", {"state": "STAGED_IN"})
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})
    events = auth_client.get("/events/")
    assert events["count"] == 10 * 3


def test_aggregated_state_history_by_batch_job(auth_client, create_session, job_dict):
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(2)])
    auth_client.bulk_put("/jobs/", {"state": "STAGED_IN"})
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"})

    session = create_session()
    session2 = create_session()

    # Batch Job 1 started and acquired all jobs:
    acquired = auth_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=40,
        filter_tags={},
        max_num_jobs=2,
        max_nodes_per_job=8,
        check=status.HTTP_200_OK,
    )
    assert len(acquired) == 2
    for job in acquired:
        assert job["batch_job_id"] == session.batch_job_id

    # Status updates:
    auth_client.put(f"/batch-jobs/{session.batch_job_id}", scheduler_id=31415, status="running")
    auth_client.bulk_put("/jobs/", {"state": "RUNNING"})

    # Can look up events by scheduler id
    events = auth_client.get("/events/", scheduler_id=31415)
    assert events["count"] == 2 * 4

    # Or by batch_job_id
    events = auth_client.get("/events/", batch_job_id=session.batch_job_id)
    print(*[(e["job_id"], e["to_state"]) for e in events["results"]], sep="\n")
    print(len(events["results"]), "events received")
    assert events["count"] == 2 * 4

    # Batch Job 2 has no associated events
    events2 = auth_client.get("/events/", batch_job_id=session2.id)
    assert events2["count"] == 0


def test_aggregated_state_history_by_tags(auth_client, job_dict):
    specs = [job_dict(tags={"foo": f"x{i}"}) for i in range(3)]
    jobs = auth_client.bulk_post("/jobs/", specs)

    for i in range(3):
        events = auth_client.get("/events/", tags=f"foo:x{i}")
        assert events["count"] == 1
        for event in events["results"]:
            assert event["job_id"] == jobs[i]["id"]

    events = auth_client.get("/events/")
    assert events["count"] == len(events["results"]) == 3


def test_aggregated_state_history_by_date_range(auth_client, job_dict):
    before_create = datetime.utcnow()
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(3)])
    time.sleep(0.1)
    after_create = datetime.utcnow()

    events = auth_client.get("/events/", timestamp_before=before_create)
    assert events["count"] == 0

    events = auth_client.get("/events/", timestamp_after=after_create)
    assert events["count"] == 0

    events = auth_client.get("/events/", timestamp_after=before_create, timestamp_before=after_create)
    assert events["count"] == 3


def test_can_delete_job(auth_client, job_dict):
    # Create 4 jobs: 2 PREPROCESSED, 2 STAGED_IN
    jobs = auth_client.bulk_post("/jobs/", [job_dict() for _ in range(4)])
    auth_client.bulk_put("/jobs/", {"state": "STAGED_IN"}, id=[j["id"] for j in jobs[:2]])
    auth_client.bulk_put("/jobs/", {"state": "PREPROCESSED"}, id=[j["id"] for j in jobs[2:]])
    # Bulk-Delete the STAGED_IN
    auth_client.bulk_delete("/jobs/", state="STAGED_IN")
    jobs = auth_client.get("/jobs/")
    assert jobs["count"] == 2
    for job in jobs["results"]:
        assert job["state"] == "PREPROCESSED"


def test_can_do_multiple_lookup_by_pk(auth_client, linear_dag):
    A, B, C = linear_dag
    ids = [j["id"] for j in (A, C)]
    jobs = auth_client.get("/jobs", id=ids)
    assert jobs["count"] == len(jobs["results"]) == 2


def test_can_traverse_dag(auth_client, linear_dag):
    A, B, C = linear_dag

    child_of_A = auth_client.get("/jobs", parent_id=A["id"])
    assert child_of_A["results"][0]["id"] == B["id"]

    child_of_B = auth_client.get("/jobs", parent_id=B["id"])
    assert child_of_B["results"][0]["id"] == C["id"]


def test_delete_recursively_deletes_children(auth_client, linear_dag, db_session):
    A, B, C = linear_dag
    assert db_session.query(models.Job).count() == 3
    auth_client.delete(f"/jobs/{A['id']}")
    db_session.expire_all()
    assert db_session.query(models.Job).count() == 0


def test_delete_recursively_deletes_children2(auth_client, linear_dag, db_session):
    A, B, C = linear_dag
    assert db_session.query(models.Job).count() == 3
    auth_client.delete(f"/jobs/{B['id']}")
    db_session.expire_all()
    assert db_session.query(models.Job).count() == 1


def test_cannot_acquire_with_another_lock_id(auth_client, create_session, job_dict, fastapi_user_test_client):
    """Passing a lock id that belongs to another user results in acquire() error"""
    # self.user (via self.client) has 10 jobs
    session = create_session()
    auth_client.bulk_post("/jobs/", [job_dict() for _ in range(3)])
    assert auth_client.get("/jobs")["count"] == 3

    other_client = fastapi_user_test_client()
    assert other_client.get("/jobs")["count"] == 0
    other_client.post(
        f"/sessions/{session.id}",
        max_wall_time_min=120,
        filter_tags={},
        max_num_jobs=5,
        max_nodes_per_job=32,
        check=status.HTTP_404_NOT_FOUND,
    )


def test_filter_transfers_by_state(site, auth_client, job_dict):
    """Can filter TransferItems by state list"""
    app = create_app(
        auth_client,
        site["id"],
        transfers={
            "hello-input": {
                "required": False,
                "direction": "in",
                "local_path": "hello.yml",
                "description": "Input file for SayHello",
            },
            "hello-output": {
                "required": True,
                "direction": "out",
                "local_path": "hello.results.xml",
                "description": "Results of SayHello",
            },
        },
    )
    job1 = job_dict(
        app_id=app["id"],
        transfers={
            "hello-input": {
                "location_alias": "MyCluster",
                "path": "/path/to/input.dat",
            },
            "hello-output": {
                "location_alias": "MyCluster",
                "path": "/path/to/result.xml",
            },
        },
    )
    job2 = job_dict(
        app_id=app["id"],
        transfers={
            "hello-output": {
                "location_alias": "MyCluster",
                "path": "/path/to/result.xml",
            }
        },
    )
    job1, job2 = auth_client.bulk_post("/jobs/", [job1, job2])
    assert job1["state"] == "READY"
    assert job2["state"] == "STAGED_IN"

    transfers = auth_client.get("/transfers", state="pending")
    assert transfers["count"] == 1

    transfers = auth_client.get("/transfers", state=["pending"])
    assert transfers["count"] == 1

    transfers = auth_client.get("/transfers", state=["awaiting_job"])
    assert transfers["count"] == 2

    transfers = auth_client.get("/transfers", state=["done", "error"])
    assert transfers["count"] == 0

    transfers = auth_client.get("/transfers", state=["awaiting_job", "pending"])
    assert transfers["count"] == 3


def test_postprocessed_job_with_stage_outs(site, auth_client, job_dict):
    """POSTPROCESSED job is acquired for stage-out before marking FINISHED"""
    app = create_app(
        auth_client,
        site["id"],
        transfers={
            "hello-input": {
                "required": False,
                "direction": "in",
                "local_path": "hello.yml",
                "description": "Input file for SayHello",
            },
            "hello-output": {
                "required": True,
                "direction": "out",
                "local_path": "hello.results.xml",
                "description": "Results of SayHello",
            },
        },
    )
    job1 = job_dict(
        app_id=app["id"],
        transfers={
            "hello-input": {
                "location_alias": "MyCluster",
                "path": "/path/to/input.dat",
            },
            "hello-output": {
                "location_alias": "MyCluster",
                "path": "/path/to/result.xml",
            },
        },
    )
    job2 = job_dict(
        app_id=app["id"],
        transfers={
            "hello-output": {
                "location_alias": "MyCluster",
                "path": "/path/to/result.xml",
            }
        },
    )
    job1, job2 = auth_client.bulk_post("/jobs/", [job1, job2])
    assert job1["state"] == "READY"
    assert job2["state"] == "STAGED_IN"

    job1 = auth_client.put(f"/jobs/{job1['id']}", state="POSTPROCESSED")
    assert job1["state"] == "POSTPROCESSED"
    job1 = auth_client.put(f"/jobs/{job1['id']}", state="STAGED_OUT")
    assert job1["state"] == "JOB_FINISHED"

    job2 = auth_client.put(f"/jobs/{job2['id']}", state="POSTPROCESSED")
    assert job2["state"] == "POSTPROCESSED"
    job2 = auth_client.put(f"/jobs/{job2['id']}", state="STAGED_OUT")
    assert job2["state"] == "JOB_FINISHED"


def test_reset_job_with_finished_parents(auth_client, linear_dag):
    A, B, C = linear_dag
    assert B["state"] == "AWAITING_PARENTS"

    A = auth_client.put(f"jobs/{A['id']}", state="POSTPROCESSED")
    assert A["state"] == "JOB_FINISHED"

    B = auth_client.get(f"jobs/{B['id']}")
    assert B["state"] == "READY"
    B = auth_client.put(f"jobs/{B['id']}", state="FAILED")
    B = auth_client.put(f"jobs/{B['id']}", state="RESET")
    assert B["state"] == "READY"



import json
import logging
import os
import re
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from dateutil.parser import ParserError

from balsam.util import parse_to_utc

from .scheduler import (
    SchedulerBackfillWindow,
    SchedulerJobLog,
    SchedulerJobStatus,
    SchedulerNonZeroReturnCode,
    SchedulerSubmitError,
    SubprocessSchedulerInterface,
    scheduler_subproc,
)

PathLike = Union[Path, str]
logger = logging.getLogger(__name__)


class LsfScheduler(SubprocessSchedulerInterface):
    status_exe = "bjobs"
    submit_exe = "bsub"
    delete_exe = "bkill"
    backfill_exe = "bslots"
    default_submit_kwargs: Dict[str, str] = {}
    submit_kwargs_flag_map: Dict[str, str] = {}

    _queue_name = "batch"

    # maps scheduler states to Balsam states
    _job_states = {
        "PEND": "queued",
        "RUN": "running",
        "BLOCKED": "failed",
    }

    @staticmethod
    def _job_state_map(scheduler_state: str) -> str:
        return LsfScheduler._job_states.get(scheduler_state, "unknown")

    # maps Balsam status fields to the scheduler fields
    # should be a comprehensive list of scheduler status fields
    _status_fields = {
        "scheduler_id": "JOBID",
        "state": "STAT",
        "queue": "QUEUE",
        "num_nodes": "NREQ_SLOT",
        "wall_time_min": "RUNTIMELIMIT",
        "project": "PROJ_NAME",
        "time_remaining_min": "RUN_TIME",
        "queued_time_min": "PEND_TIME",
    }

    @staticmethod
    def _get_envs() -> Dict[str, str]:
        env = {}
        fields = LsfScheduler._status_fields.values()
        env["LSB_BJOBS_FORMAT"] = " ".join(fields)
        return env

    # when reading these fields from the scheduler apply
    # these maps to the string extracted from the output
    @staticmethod
    def _status_field_map(balsam_field: str) -> Optional[Callable[[str], Any]]:
        status_field_map = {
            "scheduler_id": lambda id: int(id),
            "state": lambda state: LsfScheduler._job_states[state],
            "queue": lambda queue: str(queue),
            "num_nodes": lambda n: 0 if n == "-" else int(int(n) // 42),
            "wall_time_min": lambda minutes: int(float(minutes)),
            "project": lambda project: str(project),
            "time_remaining_min": lambda time: int(int(time.split()[0]) / 60),
            "queued_time_min": lambda minutes: int(minutes),
        }
        return status_field_map.get(balsam_field, None)

    @staticmethod
    def _render_submit_args(
        script_path: PathLike, project: str, queue: str, num_nodes: int, wall_time_min: int, **kwargs: Any
    ) -> List[str]:
        args = [
            LsfScheduler.submit_exe,
            "-o",
            os.path.basename(os.path.splitext(script_path)[0]) + ".output",
            "-e",
            os.path.basename(os.path.splitext(script_path)[0]) + ".error",
            "-P",
            project,
            "-q",
            queue,
            "-nnodes",
            str(int(num_nodes)),
            "-W",
            str(int(wall_time_min)),
        ]
        # adding additional flags as needed, e.g. `-C knl`
        for key, default_value in LsfScheduler.default_submit_kwargs.items():
            flag = LsfScheduler.submit_kwargs_flag_map[key]
            value = kwargs.setdefault(key, default_value)
            args += [flag, value]

        args.append(str(script_path))
        return args

    @staticmethod
    def _render_status_args(
        project: Optional[str] = None, user: Optional[str] = None, queue: Optional[str] = None
    ) -> List[str]:
        os.environ.update(LsfScheduler._get_envs())
        args = [LsfScheduler.status_exe]
        if user is not None:
            args += ["-u", user]
        if project is not None:
            pass  # not supported
        if queue is not None:
            args += ["-q", queue]
        # format output as json
        args += ["-json"]
        return args

    @staticmethod
    def _render_delete_args(job_id: Union[int, str]) -> List[str]:
        return [LsfScheduler.delete_exe, str(job_id)]

    @staticmethod
    def _render_backfill_args() -> List[str]:
        return [LsfScheduler.backfill_exe, '-R"select[CN]"']

    @classmethod
    def get_backfill_windows(cls) -> Dict[str, List[SchedulerBackfillWindow]]:
        backfill_args = cls._render_backfill_args()
        try:
            stdout = scheduler_subproc(backfill_args)
        except SchedulerNonZeroReturnCode as e:
            if "No backfill window meets" in str(e):
                return {LsfScheduler._queue_name: []}
            raise
        backfill_windows = cls._parse_backfill_output(stdout)
        return backfill_windows

    @staticmethod
    def _parse_submit_output(submit_output: str) -> int:
        if "Job not submitted" in submit_output:
            raise SchedulerSubmitError(f"job submission failed with message: {submit_output}")
        try:
            start = len("Job <")
            end = submit_output.find(">", start)
            scheduler_id = int(submit_output[start:end])
        except ValueError:
            scheduler_id = int(submit_output.split()[-1])
        return scheduler_id

    @staticmethod
    def _parse_status_output(raw_output: str) -> Dict[int, SchedulerJobStatus]:
        # Example output:
        # {
        # "COMMAND":"bjobs",
        # "JOBS":47,
        # "RECORDS":[
        #   {
        #     "JOBID":"806290",
        #     "STAT":"RUN",
        #     "QUEUE":"batch",
        #     "PROJ_NAME":"BIP152",
        #     "PEND_TIME":"17",
        #     "NREQ_SLOT":"43",
        #     "RUNTIMELIMIT":"1440.0",
        #     "RUN_TIME":"16038 second(s)"
        #   },
        json_output = json.loads(raw_output)
        status_dict = {}
        batch_jobs = json_output["RECORDS"]
        for job_data in batch_jobs:
            status = {}
            try:
                for balsam_key, scheduler_key in LsfScheduler._status_fields.items():
                    func = LsfScheduler._status_field_map(balsam_key)
                    if callable(func):
                        status[balsam_key] = func(job_data[scheduler_key])
            except KeyError:
                logger.warning("failed parsing job data: %s", job_data)
            else:
                job_stat = SchedulerJobStatus(**status)
                status_dict[job_stat.scheduler_id] = job_stat

        return status_dict

    @staticmethod
    def _parse_backfill_output(stdout: str) -> Dict[str, List[SchedulerBackfillWindow]]:
        raw_lines = stdout.split("\n")
        windows: Dict[str, List[SchedulerBackfillWindow]] = {LsfScheduler._queue_name: []}
        node_lines = raw_lines[1:]
        for line in node_lines:
            if len(line.strip()) == 0:
                continue
            windows[LsfScheduler._queue_name].append(LsfScheduler._parse_bslots_line(line))
        return windows

    @staticmethod
    def _parse_bslots_line(line: str) -> SchedulerBackfillWindow:
        parts = line.split()
        nodes = int(parts[0])
        backfill_time = 0
        if len(re.findall("hours.*minutes.*seconds", line)) > 0:
            backfill_time += int(parts[1]) * 60
            backfill_time += int(parts[3])
        elif len(re.findall("minutes.*seconds", line)) > 0:
            backfill_time += int(parts[1])

        return SchedulerBackfillWindow(num_nodes=nodes, wall_time_min=backfill_time)

    @staticmethod
    def _parse_logs(scheduler_id: Union[int, str], job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        # TODO: Return job start/stop time from log file or command
        args = [LsfScheduler.status_exe]
        args += ["-o", "JOBID START_TIME FINISH_TIME"]
        # format output as json
        args += ["-json"]
        args += [str(scheduler_id)]
        stdout = scheduler_subproc(args)
        json_output = json.loads(stdout)
        if json_output["JOBS"] == 0:
            logger.error("no job found for JOB ID = %s", scheduler_id)
            return SchedulerJobLog()
        if json_output["JOBS"] > 1:
            logger.error("something strange happened, more than one job returned: \n %s", stdout)
            return SchedulerJobLog()

        job_data = json_output["RECORDS"][0]
        start_raw = job_data.get("START_TIME")
        end_raw = job_data.get("FINISH_TIME")

        if not (start_raw and end_raw):
            logger.warning(f"parse_logs got START_TIME: {start_raw}; FINISH_TIME: {end_raw}")
            return SchedulerJobLog()
        try:
            start = parse_to_utc(start_raw, local_zone="ET")
            end = parse_to_utc(end_raw, local_zone="ET")
        except ParserError:
            logger.warning(f"Failed to parse job_data times (START_TIME: {start_raw}) (FINISH_TIME: {end_raw})")
            return SchedulerJobLog()
        return SchedulerJobLog(start_time=start, end_time=end)



import getpass
import logging
import subprocess
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import psutil  # type: ignore

from .scheduler import SchedulerBackfillWindow, SchedulerInterface, SchedulerJobLog, SchedulerJobStatus

logger = logging.getLogger(__name__)
PathLike = Union[str, Path]


class LocalProcessScheduler(SchedulerInterface):
    _state_map = {
        psutil.STATUS_ZOMBIE: "finished",
        psutil.STATUS_DEAD: "finished",
    }

    _subprocesses: List["subprocess.Popen[bytes]"] = []

    @classmethod
    def submit(
        cls,
        script_path: PathLike,
        project: str,
        queue: str,
        num_nodes: int,
        wall_time_min: int,
        cwd: Optional[PathLike] = None,
        **kwargs: Any,
    ) -> int:
        """
        Submit the script at `script_path` to a local job queue.
        Returns scheduler ID of the submitted job.
        """
        outfile = Path(script_path).resolve().with_suffix(".output")
        with open(outfile, "wb") as fp:
            p = subprocess.Popen(
                f"bash {script_path}",
                shell=True,
                stdout=fp,
                stderr=subprocess.STDOUT,
            )
        cls._subprocesses.append(p)
        return p.pid

    @classmethod
    def get_statuses(
        cls,
        project: Optional[str] = None,
        user: Optional[str] = getpass.getuser(),
        queue: Optional[str] = None,
    ) -> Dict[int, SchedulerJobStatus]:
        """
        Returns dictionary keyed on scheduler job id and a value of JobStatus for each
          job belonging to current user, project, and/or queue
        """
        results = {}
        for p in psutil.process_iter(attrs=["pid", "username", "status"]):
            if p.info["username"] == user:
                pid = int(p.info["pid"])
                status = p.info["status"]
                results[pid] = SchedulerJobStatus(
                    scheduler_id=pid,
                    state=cls._state_map.get(status, "running"),
                    queue="local",
                    num_nodes=1,
                    wall_time_min=0,
                    project="local",
                    time_remaining_min=1000,
                    queued_time_min=0,
                )
        cls._subprocesses = [p for p in cls._subprocesses if p.poll() is None]
        return results

    @classmethod
    def delete_job(cls, scheduler_id: int) -> str:
        """
        Deletes the batch job matching `scheduler_id`
        """
        logger.info(f"Delete Job {scheduler_id}")
        if not psutil.pid_exists(scheduler_id):
            logger.info(f"Delete Job could not find local PID {scheduler_id}")
            return str(scheduler_id)

        parent = psutil.Process(scheduler_id)
        processes = parent.children(recursive=True)
        processes.append(parent)

        for child in processes:
            logger.debug(f"SIGTERM process {child.pid}")
            child.terminate()

        (gone, alive) = psutil.wait_procs(processes, timeout=10.0)
        for child in alive:
            pid = child.pid
            try:
                child.kill()
                logger.warning(f"Force-killed unresponsive PID {pid}: {child.name()} {child.cmdline()}")
            except (psutil.NoSuchProcess, psutil.ZombieProcess, psutil.AccessDenied, psutil.ProcessLookupError):
                logger.warning(f"Failed to kill zombie/nonexistant PID {pid}")
        return str(scheduler_id)

    @classmethod
    def get_backfill_windows(cls) -> Dict[str, List[SchedulerBackfillWindow]]:
        """
        Returns a dictionary keyed on queue name and a value of list of
          BackfillWindow on the system for available scheduling windows
        """
        return {}

    @classmethod
    def parse_logs(cls, scheduler_id: int, job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        """
        Reads the scheduler logs to determine job metadata like start_time and end_time
        """
        return SchedulerJobLog()

    @classmethod
    def discover_projects(cls) -> List[str]:
        return ["local"]



import logging
import logging.handlers
import os
from typing import Optional, TextIO

from gunicorn import util  # type: ignore
from gunicorn.glogging import Logger  # type: ignore


class RotatingGunicornLogger(Logger):  # type: ignore
    def _set_handler(
        self, log: logging.Logger, output: str, fmt: logging.Formatter, stream: Optional[TextIO] = None
    ) -> None:
        # remove previous gunicorn log handler
        h = self._get_gunicorn_handler(log)
        if h:
            log.handlers.remove(h)

        if output is not None:
            if output == "-":
                h = logging.StreamHandler(stream)
            else:
                util.check_is_writeable(output)
                h = logging.handlers.RotatingFileHandler(
                    output,
                    maxBytes=int(32 * 1e6),
                    backupCount=3,
                )
                # make sure the user can reopen the file
                try:
                    os.chown(h.baseFilename, self.cfg.user, self.cfg.group)
                except OSError:
                    # it's probably OK there, we assume the user has given
                    # /dev/null as a parameter.
                    pass

            h.setFormatter(fmt)
            h._gunicorn = True
            log.addHandler(h)
